from pykep.core import epoch, DAY2SEC, MU_SUN, lambert_problem, propagate_lagrangian, fb_prop, AU, epoch
from pykep.planet import jpl_lp
from math import pi, cos, sin, acos, log, sqrt
import numpy as np

# Avoiding scipy dependency
def norm(x):
    return sqrt(sum([it * it for it in x]))

class mga_1dsm:
    r"""
    This class transcribes a Multiple Gravity Assist trajectory with one deep space manouvre per leg (MGA-1DSM) into an optimisation problem.
    It may be used as a User Defined Problem (UDP) for the pygmo (http://esa.github.io/pygmo/) optimisation suite.

    - Izzo, Dario. "Global optimization and space pruning for spacecraft trajectory design." Spacecraft Trajectory Optimization 1 (2010): 178-200.

    The decision vector (chromosome) is::

      direct encoding: [t0] + [u, v, Vinf, eta1, T1] + [beta, rp/rV, eta2, T2] + ... 
      alpha encoding:  [t0] + [u, v, Vinf, eta1, a1] + [beta, rp/rV, eta2, a2] + ... + [T]
      eta encoding:    [t0] + [u, v, Vinf, eta1, n1] + [beta, rp/rV, eta2, n2] + ...

      where t0 is a mjd2000, Vinf is in km/s, T in days, beta in radians and the rest non dimensional.

    .. note::

       The time of flights of a MGA-1DSM trajectory (and in general) can be encoded in different ways.
       When they are directly present in the decision vector, we have a *direct* encoding. This is the most 'evolvable' encoding
       but also the one that requires the most problem knowledge (e.g. to define the bounds on each leg) and is not 
       very flexible in dealing with constraints on the total time of flight. The *alpha* and *eta* encodings, instead, allow
       to only specify bounds on the time of flight of the entire trajectory, and not on the single legs: a property that is attractive
       for multi-objective optimization, for example.

       In the *alpha* encoding each leg time-of-flight is decoded as follows, T_i = T log(alpha_i) / \sum_n(log(alpha_n)).
       In the *eta* encoding  each leg time-of-flight is decoded as follows, T_i = (tof_max - \sum_0^(i-1)(T_j)) * eta_i

       The chromosome dimension for the direct and eta encoding is the same, while the alpha encoding requires one more gene.

    .. note::

       The resulting problem is box-bounded (unconstrained).
    """

    def __init__(self,
                 seq = [jpl_lp('earth'), jpl_lp('venus'), jpl_lp('earth')],
                 t0 = [epoch(0), epoch(1000)],
                 tof = [[10, 300], [10, 300]],
                 vinf = [0.5, 2.5],
                 add_vinf_dep = False,
                 add_vinf_arr = True,
                 tof_encoding = 'direct',
                 multi_objective = False,
                 orbit_insertion = False,
                 e_target = None,
                 rp_target = None,
                 eta_lb = 0.1,
                 eta_ub = 0.9,
                 rp_ub = 30
                 ):
        """
        pykep.trajopt.mga_1dsm(seq = [jpl_lp('earth'), jpl_lp('venus'), jpl_lp('earth')], t0 = [epoch(0),epoch(1000)], tof = [1.0,5.0], vinf = [0.5, 2.5], multi_objective = False, add_vinf_dep = False, add_vinf_arr=True)

        - seq (``list`` of ``pykep.planet``): the encounter sequence (including the starting launch)
        - t0 (``list`` of ``pykep.epoch`` or ``floats``): the launch window (in mjd2000 if floats)
        - tof (``list`` or ``float``): bounds on the time of flight (days). If *tof_encoding* is 'direct', this contains a list
            of 2D lists defining the upper and lower bounds on each leg. If *tof_encoding* is 'alpha',
            this contains a list of two floats containing the lower and upper bounds on the time-of-flight. If *tof_encoding*
            is 'eta' tof is a float defining the upper bound on the time-of-flight
        - vinf (``list``): the minimum and maximum allowed initial hyperbolic velocity (at launch), in km/sec
        - add_vinf_dep (``bool``): when True the computed Dv includes the initial hyperbolic velocity (at launch)
        - add_vinf_arr (``bool``): when True the computed Dv includes the final hyperbolic velocity (at the last planet)
        - tof_encoding (``str``): one of 'direct', 'alpha' or 'eta'. Selects the encoding for the time of flights
        - multi_objective (``bool``): when True constructs a multiobjective problem (dv, T)
        - orbit_insertion (``bool``): when True the arrival dv is computed as that required to acquire a target orbit defined by e_target and rp_target
        - e_target (``float``): if orbit_insertion is True this defines the target orbit eccentricity around the final planet
        - rp_target (``float``): if orbit_insertion is True this defines the target orbit pericenter around the final planet (in m)
        """

        # Sanity checks
        # 1 - Planets need to have the same mu_central_body
        if ([r.mu_central_body for r in seq].count(seq[0].mu_central_body) != len(seq)):
            raise ValueError(
                'All planets in the sequence need to have identical mu_central_body')
        # 2 - tof encoding needs to be one of 'alpha', 'eta', 'direct'
        if tof_encoding not in ['alpha', 'eta', 'direct']:
            raise TypeError(
                'tof encoding must be one of \'alpha\', \'eta\', \'direct\'')
        # 3 - tof is expected to have different content depending on the tof_encoding
        if tof_encoding == 'direct':
            if np.shape(np.array(tof)) != (len(seq) - 1, 2):
                raise TypeError(
                    'tof_encoding is ' + tof_encoding + ' and tof must be a list of two dimensional lists and with length equal to the number of legs')
        if tof_encoding == 'alpha':
            if np.shape(np.array(tof)) != (2,):
                raise TypeError(
                    'tof_encoding is ' + tof_encoding + ' and tof must be a list of two floats')
        if tof_encoding == 'eta':
            if np.shape(np.array(tof)) != ():
                raise TypeError(
                    'tof_encoding is ' + tof_encoding + ' and tof must be a float')
        # 4 - Check launch window t0. If defined in terms of floats transform into epochs
        if len(t0) != 2:
            raise TypeError(
                    't0 is ' + t0 + ' while should be a list of two floats or epochs')
        if type(t0[0]) is not epoch:
            t0[0] = epoch(t0[0])
        if type(t0[1]) is not epoch:
            t0[1] = epoch(t0[1])
        # 5 - Check that if orbit insertion is selected e_target and r_p are
        # defined
        if orbit_insertion:
            if rp_target is None:
                raise ValueError(
                    'The rp_target needs to be specified when orbit insertion is selected')
            if e_target is None:
                raise ValueError(
                    'The e_target needs to be specified when orbit insertion is selected')
            if add_vinf_arr is False:
                raise ValueError(
                    'When orbit insertion is selected, the add_vinf_arr must be True')

        self._seq = seq
        self._t0 = t0
        self._tof = tof
        self._vinf = vinf
        self._add_vinf_dep = add_vinf_dep
        self._add_vinf_arr = add_vinf_arr
        self._tof_encoding = tof_encoding
        self._multi_objective = multi_objective
        self._orbit_insertion = orbit_insertion
        self._e_target = e_target
        self._rp_target = rp_target
        self._eta_lb = eta_lb
        self._eta_ub = eta_ub
        self._rp_ub = rp_ub

        self.n_legs = len(seq) - 1
        self.common_mu = seq[0].mu_central_body

    def get_nobj(self):
        return self._multi_objective + 1

    def get_bounds(self):
        t0 = self._t0
        tof = self._tof
        vinf = self._vinf
        seq = self._seq
        # Base for all possiblities (eta encoding)
        lb = [t0[0].mjd2000] + [0.0, 0.0, vinf[0] * 1000, self._eta_lb,
                                1e-3] + [-2 * pi, np.nan, self._eta_lb, 1e-3] * (self.n_legs - 1)
        ub = [t0[1].mjd2000] + [1.0, 1.0, vinf[1] * 1000, self._eta_ub,
                                1.0 - 1e-3] + [2 * pi, self._rp_ub, self._eta_ub, 1.0 - 1e-3] * (self.n_legs - 1)
        # Distinguishing among cases (only direct and alpha)
        if self._tof_encoding == 'alpha':
            lb = lb + [tof[0]]
            ub = ub + [tof[1]]
        elif self._tof_encoding == 'direct':
            for i in range(self.n_legs):
                lb[5 + 4 * i] = tof[i][0]
                ub[5 + 4 * i] = tof[i][1]

        # Setting the minimum rp/rP using the planet safe radius
        for i, pl in enumerate(seq[1:-1]):
            lb[7 + 4 * i] = pl.safe_radius / pl.radius
        return (lb, ub)

    def _decode_times_and_vinf(self, x):
        # 1 - we decode the times of flight
        if self._tof_encoding == 'alpha':
            # decision vector is  [t0] + [u, v, Vinf, eta1, a1] + [beta, rp/rV, eta2, a2] + ... + [T]
            T = list([0] * (self.n_legs))
            for i in range(len(T)):
                T[i] = -log(x[5 + 4 * i])
            alpha_sum = sum(T)
            retval_T = [x[-1] * time / alpha_sum for time in T]
        elif self._tof_encoding == 'direct':
            # decision vector is  [t0] + [u, v, Vinf, eta1, T1] + [beta, rp/rV, eta2, T2] + ...
            retval_T = x[5::4]
        elif self._tof_encoding == 'eta':
            # decision vector is  [t0] + [u, v, Vinf, eta1, n1] + [beta, rp/rV, eta2, n2] + ...
            dt = self._tof
            T = [0] * self.n_legs
            for i in range(self.n_legs):
                T[i] = (dt - sum(T[:i])) * x[5 + 4*i]
            retval_T = T

        # 2 - we decode the hyperbolic velocity at departure
        theta = 2 * pi * x[1]
        phi = acos(2 * x[2] - 1) - pi / 2

        Vinfx = x[3] * cos(phi) * cos(theta)
        Vinfy = x[3] * cos(phi) * sin(theta)
        Vinfz = x[3] * sin(phi)

        return (retval_T, Vinfx, Vinfy, Vinfz)

    # Objective function
    def fitness(self, x):
        # 1 -  we 'decode' the chromosome recording the various times of flight
        # (days) in the list T and the cartesian components of vinf
        T, Vinfx, Vinfy, Vinfz = self._decode_times_and_vinf(x)

        # 2 - We compute the epochs and ephemerides of the planetary encounters
        t_P = list([None] * (self.n_legs + 1))
        r_P = list([None] * (self.n_legs + 1))
        v_P = list([None] * (self.n_legs + 1))
        DV = list([0.0] * (self.n_legs + 1))
        for i in range(len(self._seq)):
            t_P[i] = epoch(x[0] + sum(T[0:i]))
            r_P[i], v_P[i] = self._seq[i].eph(t_P[i])

        # 3 - We start with the first leg
        v0 = [a + b for a, b in zip(v_P[0], [Vinfx, Vinfy, Vinfz])]
        r, v = propagate_lagrangian(
            r_P[0], v0, x[4] * T[0] * DAY2SEC, self.common_mu)

        # Lambert arc to reach seq[1]
        dt = (1 - x[4]) * T[0] * DAY2SEC
        l = lambert_problem(
            r, r_P[1], dt, self.common_mu, cw = False, max_revs = 0)
        v_end_l = l.get_v2()[0]
        v_beg_l = l.get_v1()[0]

        # First DSM occuring at time nu1*T1
        DV[0] = norm([a - b for a, b in zip(v_beg_l, v)])

        # 4 - And we proceed with each successive leg
        for i in range(1, self.n_legs):
            # Fly-by
            v_out = fb_prop(v_end_l, v_P[i], x[
                            7 + (i - 1) * 4] * self._seq[i].radius, x[6 + (i - 1) * 4], self._seq[i].mu_self)
            # s/c propagation before the DSM
            r, v = propagate_lagrangian(
                r_P[i], v_out, x[8 + (i - 1) * 4] * T[i] * DAY2SEC, self.common_mu)
            # Lambert arc to reach Earth during (1-nu2)*T2 (second segment)
            dt = (1 - x[8 + (i - 1) * 4]) * T[i] * DAY2SEC
            l = lambert_problem(r, r_P[i + 1], dt,
                                self.common_mu, cw=False, max_revs=0)
            v_end_l = l.get_v2()[0]
            v_beg_l = l.get_v1()[0]
            # DSM occuring at time nu2*T2
            DV[i] = norm([a - b for a, b in zip(v_beg_l, v)])

        # Last Delta-v
        if self._add_vinf_arr:
            DV[-1] = norm([a - b for a, b in zip(v_end_l, v_P[-1])])
            if self._orbit_insertion:
                # In this case we compute the insertion DV as a single pericenter
                # burn
                DVper = np.sqrt(DV[-1] * DV[-1] + 2 *
                                self._seq[-1].mu_self / self._rp_target)
                DVper2 = np.sqrt(2 * self._seq[-1].mu_self / self._rp_target -
                                self._seq[-1].mu_self / self._rp_target * (1. - self._e_target))
                DV[-1] = np.abs(DVper - DVper2)

        if self._add_vinf_dep:
            DV[0] += x[3]

        if not self._multi_objective:
            return (sum(DV),)
        else:
            return (sum(DV), sum(T))

    def pretty(self, x):
        """
        prob.plot(x)

        - x: encoded trajectory

        Prints human readable information on the trajectory represented by the decision vector x

        Example::

          print(prob.pretty(x))
        """
       # 1 -  we 'decode' the chromosome recording the various times of flight
        # (days) in the list T and the cartesian components of vinf
        T, Vinfx, Vinfy, Vinfz = self._decode_times_and_vinf(x)

        # 2 - We compute the epochs and ephemerides of the planetary encounters
        t_P = list([None] * (self.n_legs + 1))
        r_P = list([None] * (self.n_legs + 1))
        v_P = list([None] * (self.n_legs + 1))
        DV = list([0.0] * (self.n_legs + 1))
        for i in range(len(self._seq)):
            t_P[i] = epoch(x[0] + sum(T[0:i]))
            r_P[i], v_P[i] = self._seq[i].eph(t_P[i])

        # 3 - We start with the first leg
        print("First Leg: " + self._seq[0].name + " to " + self._seq[1].name)
        print("Departure: " + str(t_P[0]) +
              " (" + str(t_P[0].mjd2000) + " mjd2000) ")
        print("Duration: " + str(T[0]) + "days")
        print("VINF: " + str(x[3] / 1000) + " km/sec")

        v0 = [a + b for a, b in zip(v_P[0], [Vinfx, Vinfy, Vinfz])]
        r, v = propagate_lagrangian(
            r_P[0], v0, x[4] * T[0] * DAY2SEC, self.common_mu)

        print("DSM after " + str(x[4] * T[0]) + " days")

        # Lambert arc to reach seq[1]
        dt = (1 - x[4]) * T[0] * DAY2SEC
        l = lambert_problem(
            r, r_P[1], dt, self.common_mu, cw = False, max_revs = 0)
        v_end_l = l.get_v2()[0]
        v_beg_l = l.get_v1()[0]

        # First DSM occuring at time nu1*T1
        DV[0] = norm([a - b for a, b in zip(v_beg_l, v)])
        print("DSM magnitude: " + str(DV[0]) + "m/s")

        # 4 - And we proceed with each successive leg
        for i in range(1, self.n_legs):
            print("\nleg no. " + str(i + 1) + ": " +
                  self._seq[i].name + " to " + self._seq[i + 1].name)
            print("Duration: " + str(T[i]) + "days")
            # Fly-by
            v_out = fb_prop(v_end_l, v_P[i], x[
                            7 + (i - 1) * 4] * self._seq[i].radius, x[6 + (i - 1) * 4], self._seq[i].mu_self)
            print(
                "Fly-by epoch: " + str(t_P[i]) + " (" + str(t_P[i].mjd2000) + " mjd2000) ")
            print(
                "Fly-by radius: " + str(x[7 + (i - 1) * 4]) + " planetary radii")
            # s/c propagation before the DSM
            r, v = propagate_lagrangian(
                r_P[i], v_out, x[8 + (i - 1) * 4] * T[i] * DAY2SEC, self.common_mu)
            print("DSM after " + str(x[8 + (i - 1) * 4] * T[i]) + " days")
            # Lambert arc to reach Earth during (1-nu2)*T2 (second segment)
            dt = (1 - x[8 + (i - 1) * 4]) * T[i] * DAY2SEC
            l = lambert_problem(r, r_P[i + 1], dt,
                                self.common_mu, cw=False, max_revs=0)
            v_end_l = l.get_v2()[0]
            v_beg_l = l.get_v1()[0]
            # DSM occuring at time nu2*T2
            DV[i] = norm([a - b for a, b in zip(v_beg_l, v)])
            print("DSM magnitude: " + str(DV[i]) + "m/s")

        # Last Delta-v
        print("\nArrival at " + self._seq[-1].name)
        DV[-1] = norm([a - b for a, b in zip(v_end_l, v_P[-1])])
        print(
            "Arrival epoch: " + str(t_P[-1]) + " (" + str(t_P[-1].mjd2000) + " mjd2000) ")
        print("Arrival Vinf: " + str(DV[-1]) + "m/s")
        if self._orbit_insertion:
            # In this case we compute the insertion DV as a single pericenter
            # burn
            DVper = np.sqrt(DV[-1] * DV[-1] + 2 *
                            self._seq[-1].mu_self / self._rp_target)
            DVper2 = np.sqrt(2 * self._seq[-1].mu_self / self._rp_target -
                            self._seq[-1].mu_self / self._rp_target * (1. - self._e_target))
            DVinsertion = np.abs(DVper - DVper2)
            print("Insertion DV: " + str(DVinsertion) + "m/s")

        print("Total mission time: " + str(sum(T) / 365.25) + " years (" + str(sum(T)) + " days)")

    # Plot of the trajectory
    def plot(self, x, ax = None):
        """
        ax = prob.plot(x, ax=None)

        - x: encoded trajectory
        - ax: matplotlib axis where to plot. If None figure and axis will be created
        - [out] ax: matplotlib axis where to plot

        Plots the trajectory represented by a decision vector x on the 3d axis ax

        Example::

          ax = prob.plot(x)
        """
        import matplotlib as mpl
        from mpl_toolkits.mplot3d import Axes3D
        import matplotlib.pyplot as plt
        from pykep.orbit_plots import plot_planet, plot_lambert, plot_kepler

        if ax is None:
            mpl.rcParams['legend.fontsize'] = 10
            fig = plt.figure()
            axis = fig.gca(projection='3d')
        else:
            axis = ax

        axis.scatter(0, 0, 0, color='y')

        # 1 -  we 'decode' the chromosome recording the various times of flight
        # (days) in the list T and the cartesian components of vinf
        T, Vinfx, Vinfy, Vinfz = self._decode_times_and_vinf(x)

        # 2 - We compute the epochs and ephemerides of the planetary encounters
        t_P = list([None] * (self.n_legs + 1))
        r_P = list([None] * (self.n_legs + 1))
        v_P = list([None] * (self.n_legs + 1))
        DV = list([None] * (self.n_legs + 1))

        for i, planet in enumerate(self._seq):
            t_P[i] = epoch(x[0] + sum(T[0:i]))
            r_P[i], v_P[i] = planet.eph(t_P[i])
            plot_planet(planet, t0=t_P[i], color=(
                0.8, 0.6, 0.8), legend=True, units=AU, axes=axis, N=150)

        # 3 - We start with the first leg
        v0 = [a + b for a, b in zip(v_P[0], [Vinfx, Vinfy, Vinfz])]
        r, v = propagate_lagrangian(
            r_P[0], v0, x[4] * T[0] * DAY2SEC, self.common_mu)

        plot_kepler(r_P[0], v0, x[4] * T[0] * DAY2SEC, self.common_mu,
                    N=100, color='b', legend=False, units=AU, axes=axis)

        # Lambert arc to reach seq[1]
        dt = (1 - x[4]) * T[0] * DAY2SEC
        l = lambert_problem(r, r_P[1], dt, self.common_mu, False, False)
        plot_lambert(l, sol=0, color='r', legend=False, units=AU, axes=axis)
        v_end_l = l.get_v2()[0]
        v_beg_l = l.get_v1()[0]

        # First DSM occuring at time nu1*T1
        DV[0] = norm([a - b for a, b in zip(v_beg_l, v)])

        # 4 - And we proceed with each successive leg
        for i in range(1, self.n_legs):
            # Fly-by
            v_out = fb_prop(v_end_l, v_P[i], x[
                            7 + (i - 1) * 4] * self._seq[i].radius, x[6 + (i - 1) * 4], self._seq[i].mu_self)
            # s/c propagation before the DSM
            r, v = propagate_lagrangian(
                r_P[i], v_out, x[8 + (i - 1) * 4] * T[i] * DAY2SEC, self.common_mu)
            plot_kepler(r_P[i], v_out, x[8 + (i - 1) * 4] * T[i] * DAY2SEC,
                        self.common_mu, N=100, color='b', legend=False, units=AU, axes=axis)
            # Lambert arc to reach Earth during (1-nu2)*T2 (second segment)
            dt = (1 - x[8 + (i - 1) * 4]) * T[i] * DAY2SEC

            l = lambert_problem(r, r_P[i + 1], dt,
                                self.common_mu, False, False)
            plot_lambert(l, sol=0, color='r', legend=False,
                         units=AU, N=1000, axes=axis)

            v_end_l = l.get_v2()[0]
            v_beg_l = l.get_v1()[0]
            # DSM occuring at time nu2*T2
            DV[i] = norm([a - b for a, b in zip(v_beg_l, v)])
        plt.show()
        return axis

    def get_extra_info(self):
        return ("\n\t Sequence: " + [pl.name for pl in self._seq].__repr__() +
                "\n\t Add launcher vinf to the objective?: " + self._add_vinf_dep.__repr__() +
                "\n\t Add final vinf to the objective?: " + self._add_vinf_arr.__repr__())



import numpy as np
from numba import jit


#@jit
def gravity_spherical_harmonic(x, r_planet, mu, c, s, n_max, m_max):
    """
    Calculate the gravitational acceleration due to the spherical harmonics gravity model supplied.

    Args:
        - x (``array-like``): (N x 3) array of Cartesian coordinates of satellite in frame defined by gravity model.
        - r_planet (``float``): Equatorial radius of central body.
        - mu (``float``): Gravitational parameter of central body.
        - c (``array-like``): Two-dimensional normalised C coefficient array: C[degree, order] = C_(n, m).
        - s (``array-like``): Two-dimensional normalised S coefficient array: S[degree, order] = S_(n, m).
        - n_max (``int``): Degree up to which to calculate the gravitational acceleration.
        - m_max (``int``): Order up to which to calculate the gravitational acceleration. Cannot be higher than n_max.
    
    Returns:
        - acc (``array-like``): (N x 3) array of the gravitational acceleration.

    Example::

        r, mu, c, s, n, m = pykep.util.load_gravity_model('gravity_models/Moon/glgm3150.txt')
        x = numpy.array([[459.5, 795.8773461, 1591.754692], [-459.5, -795.8773461, -1591.754692]])
        
        acc_zonal = pykep.util.gravity_spherical_harmonic(x, r, mu, c, s, 20, 0)
        acc_sqr = pykep.util.gravity_spherical_harmonic(x, r, mu, c, s, 20, 20)
        acc_high = pykep.util.gravity_spherical_harmonic(x, r, mu, c, s, 900, 900)

    .. note::

        This model was taken from a report by NASA:
        https://ntrs.nasa.gov/archive/nasa/casi.ntrs.nasa.gov/20160011252.pdf
        This is the normalised gottlieb algorithm, as coded in MATLAB in the report and transferred to Python.
    """
    if not (len(x[0]) == 3):
        raise ValueError(f"Position must be an (N x 3) array. Shape of position is ({len(x)} x {len(x[0])}).")

    r = np.min(np.linalg.norm(x, axis=1))
    if r < r_planet:
        raise ValueError(f"Radial position is less than defined radius of central body ({r} < {r_planet}).")

    if m_max > n_max:
        raise ValueError(f"Order of model is larger than degree ({m_max} > {n_max}).")

    acc = _gottlieb(x, r_planet, mu, c, s, n_max, m_max)

    return acc


def _gottlieb(x, r_planet, mu, c, s, n_max, m_max):
    norm1, norm2, norm11, normn10, norm1m, norm2m, normn1 = _calculate_normalisation_parameters(n_max)

    n_coor = x.shape[0]

    r = np.sqrt(x[:, 0]**2 + x[:, 1]**2 + x[:, 2]**2)

    r_inverted = 1 / r

    x_r = x[:, 0] * r_inverted
    y_r = x[:, 1] * r_inverted
    z_r = x[:, 2] * r_inverted

    rp_r = r_planet * r_inverted
    rp_rn = np.copy(rp_r)

    mu_r2 = mu * r_inverted * r_inverted

    shape = (max(2, n_max + 1), max(2, n_max + 2), n_coor)
    p = np.zeros(shape)

    p[0, 0] = 1
    p[1, 0, :] = np.sqrt(3) * z_r
    p[1, 1] = np.sqrt(3)

    # sectorial ALFs
    for n in range(2, n_max + 1):
        p[n, n] = norm11[n] * p[n - 1, n - 1] * (2 * n - 1)

    shape = (max(3, n_max + 1), n_coor)
    ctil = np.ones(shape)
    stil = np.zeros(shape)

    ctil[1] = x_r
    stil[1] = y_r

    sumh = np.zeros(n_coor)
    sumgm = np.ones(n_coor)
    sumj = np.zeros(n_coor)
    sumk = np.zeros(n_coor)

    for n in range(2, n_max + 1):
        rp_rn *= rp_r

        n2m1 = 2 * n - 1
        nm1 = n - 1
        nm2 = n - 2
        np1 = n + 1

        # tesseral ALFs
        p[n, nm1] = normn1[n, nm1] * z_r * p[n, n]

        # zonal ALFs
        p[n, 0] = (n2m1 * z_r * norm1[n] * p[nm1, 0] - nm1 * norm2[n] * p[nm2, 0]) / n
        p[n, 1] = (n2m1 * z_r * norm1m[n, 1] * p[nm1, 1] - n * norm2m[n, 1] * p[nm2, 1]) / nm1

        sumhn = normn10[n] * p[n, 1] * c[n, 0]
        sumgmn = p[n, 0] * c[n, 0] * np1

        if m_max > 0:
            for m in range(2, n - 1):
                p[n, m] = (n2m1 * z_r * norm1m[n, m] * p[nm1, m] -
                           (nm1 + m) * norm2m[n, m] * p[nm2, m]) / (n - m)

            sumjn = np.zeros(n_coor)
            sumkn = np.zeros(n_coor)

            ctil[n] = ctil[1] * ctil[nm1] - stil[1] * stil[nm1]
            stil[n] = stil[1] * ctil[nm1] + ctil[1] * stil[nm1]

            for m in range(1, min(n, m_max) + 1):
                mm1 = m - 1
                mp1 = m + 1
                mxpnm = m * p[n, m]

                bnmtil = c[n, m] * ctil[m] + s[n, m] * stil[m]

                sumhn += normn1[n, m] * p[n, mp1] * bnmtil
                sumgmn += (n + m + 1) * p[n, m] * bnmtil

                bnmtm1 = c[n, m] * ctil[mm1] + s[n, m] * stil[mm1]
                anmtm1 = c[n, m] * stil[mm1] - s[n, m] * ctil[mm1]

                sumjn += mxpnm * bnmtm1
                sumkn -= mxpnm * anmtm1

            sumj += rp_rn * sumjn
            sumk += rp_rn * sumkn

        sumh += rp_rn * sumhn
        sumgm += rp_rn * sumgmn

    _lambda = sumgm + z_r * sumh

    acc = np.zeros(x.shape)

    acc[:, 0] = - mu_r2 * (_lambda * x_r - sumj)
    acc[:, 1] = - mu_r2 * (_lambda * y_r - sumk)
    acc[:, 2] = - mu_r2 * (_lambda * z_r - sumh)

    return acc


@jit
def _calculate_normalisation_parameters(n_max):
    """
    Calculate the normalisation parameters for the normalised Gottlieb algorithm. The mapping from the parameters defined in the report
    (https://ntrs.nasa.gov/archive/nasa/casi.ntrs.nasa.gov/20160011252.pdf) to the names used in the code looks like

        - lambda_n-1(n)       = norm1
        - lambda_n-2(n)       = norm2
        - lambda_n-1_n-1(n)   = norm11
        - lambda_n-1_m(n, m)  = norm1m
        - lambda_n-2_m(n, m)  = norm2m
        - lambda_n_m+1(n, m)  = normn1
        - lamdba_n_m+1(n, 0)  = normn10

    Args:
        - n_max (``int``): Maximum degree.

    Returns:
        Arrays of normalisation parameters.
    """
    m_max = n_max
    n_max += 1

    norm1 = np.zeros(n_max)
    norm2 = np.zeros(n_max)
    norm11 = np.zeros(n_max)
    normn10 = np.zeros(n_max)

    norm1m = np.zeros((n_max, m_max))
    norm2m = np.zeros((n_max, m_max))
    normn1 = np.zeros((n_max, m_max + 1))

    for n in range(2, n_max):
        norm1[n] = np.sqrt((2 * n + 1) / (2 * n - 1))
        norm2[n] = np.sqrt((2 * n + 1) / (2 * n - 3))
        norm11[n] = np.sqrt((2 * n + 1) / (2 * n)) / (2 * n - 1)
        normn10[n] = np.sqrt((n + 1) * n / 2)

        for m in range(1, n):  
            norm1m[n, m] = np.sqrt((n - m) * (2 * n + 1) / ((n + m) * (2 * n - 1)))  
            norm2m[n, m] = np.sqrt((n - m) * (n - m - 1) * (2 * n + 1) / ((n + m) * (n + m - 1) * (2 * n - 3)))  
            normn1[n, m] = np.sqrt((n + m + 1) * (n - m))  

    return norm1, norm2, norm11, normn10, norm1m, norm2m, normn1



from pykep.pontryagin._dynamics import _dynamics
from pykep.core import MU_SUN, epoch, AU
from pykep.sims_flanagan import spacecraft, sc_state
from scipy.integrate import ode
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from mpl_toolkits.mplot3d import Axes3D


class leg(object):
    """Indirect optimal control transcription trajectory leg.

    This class represents an indirect optimal control transcription
    (alla moda di `Pontryagin's maximum principle <https://en.wikipedia.org/wiki/Pontryagin%27s_maximum_principle>`_) of a generic
    two-point boundary trajectory.

    Attributes:
        - t0 (``float``): Departure time [mjd2000].
        - x0 (``numpy.ndarray``): Cartesian departure state [m, m, m, m/s, m/s, m/s, kg].
        - l0 (``numpy.ndarray``): Costate variables [ND, ND, ND, ND, ND, ND, ND].
        - tf (``float``): Arrival time [mjd2000].
        - xf (``numpy.ndarray``): Cartesian arrival state [m, m, m, m/s, m/s, m/s, kg].
        - sc (``pykep.sims_flanagan.spacecraft``): Generic spacecraft with propulsive properties.
        - mu (``float``): Gravitational parametre of primary body [m^3/s^2].
        - freemass (``bool``): Activates final mass transversality condition.
        - freetime (``bool``): Activates final time transversality condition. Allows final time to vary.
        - alpha (``float``): Homotopy parametre.
        - bound (``bool``): Activates bounded control.
        - nec (``int``): Number of equality constraints.
        - trajectory (``numpy.ndarray``): Array of nondimensional fullstates.
        - times (``numpy.ndarray``): Vector of nondimensional integration times.
    """

    def __init__(self, t0=None, x0=None, l0=None, tf=None, xf=None, sc=spacecraft(1000, 0.3, 2500), mu=MU_SUN, freemass=True, freetime=True, alpha=1, bound=True):
        """Initialises an indirect optimal control transcription trajectory leg.

        Args:
            - t0 (``pykep.epoch``, ``None``): Departure time [mjd2000].
            - x0 (``pykep.sims_flanagan.sc_state``, ``None``): Cartesian departure state [m, m, m, m/s, m/s, m/s, kg].
            - l0 (``numpy.ndarray``, ``list``, ``tuple``, ``None``): Costate variables [ND, ND, ND, ND, ND, ND, ND].
            - tf (``pykep.epoch``, ``None``): Arrival time [mjd2000].
            - xf (``pykep.sims_flanagan.sc_state``, ``None``): Cartesian arrival state [m, m, m, m/s, m/s, m/s, kg].
            - sc (``pykep.sims_flanagan.spacecraft``): Generic spacecraft with propulsive properties.
            - mu (``float``, ``int``): Gravitational parametre of primary body [m^3/s^2].
            - freemass (``bool``): Activates final mass transversality condition.
            - freetime (``bool``): Activates final time transversality condition. Allows final time to vary.
            - alpha (``float``, ``int``): Homotopy parametre, governing the degree to which the theoretical control law is intended to reduce propellant expenditure or energy. Setting the parametre to 1 enforces a mass-optimal control law, with a characteristic bang-bang control profile (either full throttle or off). Setting the parametre to 0 enforces a pure quadratic control law.
            - bound (``bool``): Activates bounded control, in which the control throttle is bounded between 0 and 1, otherwise the control throttle is allowed to unbounded.

        Raises:
            - TypeError: If ``sc`` is not supplied as an instance of `pykep.sims_flanagan.spacecraft`.
            - TypeError: If ``mu`` is neither supplied as an instance of ``float`` nor ``int``.
            - ValueError: If ``mu`` is not supplied as a positive number.
            - TypeError: If either ``freemass``, ``freetime``, or ``bound`` is not supplied as an instance of ``bool``.
            - AttributeError: If equality constraint dimension cannot be determined form ``freemass`` and ``freetime``.
            - TypeError: If ``alpha`` is neither supplied as an instance of ``float`` nor ``int``.
            - ValueError: If ``alpha`` is not in between 0 and 1.
            - ValueError: If ``alpha == 1`` and ``bound == False``. Control cannot be unbounded with mass-optimal control.
            - TypeError: If either ``t0`` or ``tf`` is not supplied as an instance of ``pykep.epoch``.
            - ValueError: If departure time, ``t0.mjd2000`` does not occur before arrival time, `t`f.mjd2000``.
            - TypeError: If either ``x0`` or ``xf`` is not supplied as an instance of ``pykep.sims_flanagan.sc_state``.
            - TypeError: If costate, ``l0``, is neither supplied as an instance of ``numpy.ndarray``, ``list``, nor ``tuple``.
            - TypeError: If costate, ``l0``, is does not have 7 elements. Each element of ``l0`` corresponds to the respective elements of ``x0``.

        Examples:
            >>> l = pykep.pontryagin.leg()
            >>> l = pykep.pontryagin.leg(t0, x0, l0, tf, xf)

        .. note::
            Typically spacecraft trajectories are optimised to reduce final
            mass, which theoretically results in a bang-bang control profile.
            However, a bang-bang control profile (either 0 or 1) is problematic
            to an optimiser due to its discontinous nature. The trajectory
            optimisation problem becomes easier when one first optimises with
            unbounded quadratic control (``alpha == 0 and bound == False``),
            stores the solution, then uses the solution to optimise with
            bounded quadratic control (``alpha == 0 and bound == True``). Then
            using the solution from bounded quadratic control, one can
            incrementally optimise for increasing homotopy parametres
            (e.g. ``alphas = numpy.linspace(0, 1, 200)``) until a mass-optimal
            control solution converges.

        .. note::
            Boundary conditions ``t0``, ``x0``, ``l0``, ``tf``, and ``xf``
            are optional in the constructor. If some, but not all, boundary
            conditions are supplied, they are not set and disregarded. If the
            boundary conditions are not set upon construction (``__init__``),
            they must be set with ``set(t0, x0, l0, tf, xf)``.

        .. note::
            ``mismatch_constraints`` will not work unless the boundary
            conditions have been set either through ``__init__`` or ``set``.
        """

        # check spacecraft
        if isinstance(sc, spacecraft):
            self.spacecraft = sc
        else:
            raise TypeError(
                "Spacecraft, sc, must be of pontryagin.spacecraft type.")

        # check gravitational parametre
        if not (isinstance(mu, float) or not isinstance(mu, int)):
            raise TypeError(
                "Gravitational parametre, mu, must be supplied as float or int.")
        elif not mu > 0:
            raise ValueError("Gravitational parametre, mu, must be positive.")
        else:
            self.mu = float(mu)

        # dynamics
        self._dynamics = _dynamics(
            sc=self.spacecraft, mu=mu, alpha=alpha, bound=bound)

        # check freetime, freemass, and bound
        if not all([isinstance(param, bool) for param in [freemass, freetime, bound]]):
            raise TypeError(
                "Freemass, freetime, bound parametres must supplied as booleans.")
        else:
            self.freemass = bool(freemass)
            self.freetime = bool(freetime)
            self.bound = bool(bound)

        # equality constraint dimensionality
        if self.freemass and self.freetime:
            self.nec = 8
        elif self.freemass and not self.freetime:
            self.nec = 7
        elif not self.freemass and self.freetime:
            self.nec = 8
        elif not self.freemass and not self.freetime:
            self.nec = 7
        else:
            raise AttributeError(
                "Could not determine equality constraint dimensionality.")

        # check alpha
        if not (isinstance(alpha, float) or isinstance(alpha, int)):
            raise TypeError(
                "Homotopy parametre, alpha, must be supplied as float or int.")
        elif not (alpha >= 0 and alpha <= 1):
            raise ValueError(
                "Homotopy parametre, alpha, must be between 0 and 1.")
        elif (alpha == 1 and self.bound == False):
            raise ValueError(
                "If homotopy parametre, alpha, is 1, control must be bounded.")
        else:
            self.alpha = float(alpha)

        # if any of the necessary boundaries are not supplied
        if any(elem is None for elem in [t0, x0, l0, tf, xf]):
            pass

        # check validity if they're all supplied
        else:

            # check departure and arrival times
            if not all([isinstance(t, epoch) for t in [t0, tf]]):
                raise TypeError(
                    "Departure and arrival times, t0 & tf, must be supplied as pykep.epoch.")
            elif not t0.mjd2000 < tf.mjd2000:
                raise ValueError(
                    "Departure time must occur before arrival time.")
            else:
                self.t0 = float(t0.mjd2000)
                self.tf = float(tf.mjd2000)

            # check departure and arrival states
            states = [x0, xf]
            if not all([isinstance(state, sc_state) for state in states]):
                raise TypeError(
                    "x0 and xf must be instances of pykep.sims_flanagan.sc_state.")
            else:
                self.x0 = np.asarray(x0.get(), np.float64)
                self.xf = np.asarray(xf.get(), np.float64)

            # check departure costate
            if not (isinstance(l0, list) or isinstance(l0, np.ndarray) or isinstance(l0, tuple)):
                raise TypeError(
                    "Departure costate, l0, must be supplied as either a list, numpy.ndarray, or tuple.")
            elif len(l0) != 7:
                raise TypeError("Costate vector, l0, must be 7-dimensional.")
            else:
                self.l0 = np.asarray(l0, np.float64)

        # integrator
        self._integrator = ode(
            lambda t, fs: self._dynamics._eom_fullstate(fs),
            #lambda t, fs: self._dynamics._eom_fullstate_jac(fs)
        )

    def _recorder(self, t, fs):

        # append time
        self.times = np.append(self.times, t)

        # append fullstate
        self.trajectory = np.vstack((self.trajectory, fs))

    def set(self, t0, x0, l0, tf, xf):
        """Sets the departure and arrival boundary conditions of the trajectory.

        This instance method sets the departure time `t0`, departure state `x0`,
        departure costate `l0`, arrival time `tf`, arrival state `xf`.

        Args:
            - t0 (``pykep.epoch``): Departure time [mjd2000].
            - x0 (``pykep.sims_flanagan.sc_state``): Cartesian departure state [m, m, m, m/s, m/s, m/s, kg].
            - l0 (``numpy.ndarray``, ``list``, ``tuple``): Costate variables [ND, ND, ND, ND, ND, ND, ND].
            - tf (``pykep.epoch``): Arrival time [mjd2000].
            - xf (``pykep.epoch``): Cartesian arrival state [m, m, m, m/s, m/s, m/s, kg].

        Raises:
            - TypeError: If either ``t0`` or ``tf`` is not supplied as an instance of ``pykep.epoch``.
            - TypeError: If either ``x0`` or ``xf`` is not supllied as an instance of ``pykep.sims_flanagan.sc_state``.
            - TypeError: If ``l0`` is neither supllied as a ``numpy.ndarray``, ``list``, nor ``tuple``.

        .. note::
            If ``args`` were not supplied in ``__init__``, ``set`` must be called before calling ``mismatch_constraints``.

        Examples:
            >>> l = leg() # t0, x0, l0, tf, xf not set yet
            >>> l.set(t0, x0, l0, tf, xf) # now they're set
        """

        # check inputs
        if not all([isinstance(t, epoch) for t in [t0, tf]]):
            raise TypeError("t0 and tf must be instances of pykep.epoch.")
        elif not t0.mjd2000 < tf.mjd2000:
            raise ValueError("t0 must be less than tf.")
        elif not all([isinstance(state, sc_state) for state in [x0, xf]]):
            raise TypeError(
                "x0 and xf must be instances of pykep.sims_flanagan.sc_state.")
        elif not (isinstance(l0, tuple) or isinstance(l0, list) or isinstance(l0, np.ndarray)):
            raise TypeError(
                "l0 must be either a tuple, list, or numpy.ndarray.")
        elif len(l0) != 7:
            raise TypeError("Costate vector must be 7-dimensional.")
        else:

            # departure
            self.t0 = float(t0.mjd2000)
            self.x0 = np.asarray(x0.get(), np.float64)
            self.l0 = np.asarray(l0, np.float64)

            # arrival
            self.tf = float(tf.mjd2000)
            self.xf = np.asarray(xf.get(), np.float64)

    def _propagate(self, atol, rtol):

        # nondimensionalise departure state
        x0 = np.copy(self.x0)  # NOTE: very important
        x0[0:3] /= self._dynamics.L
        x0[3:6] /= self._dynamics.V
        x0[6] /= self._dynamics.M

        # nondimensional fullstate
        fs0 = np.hstack((x0, self.l0, [0]))

        # convert mjd2000 to mjs2000
        t0 = self.t0 * 24 * 60 * 60
        tf = self.tf * 24 * 60 * 60

        # nondimensionalise times
        t0 /= self._dynamics.T
        tf /= self._dynamics.T

        # clear trajectory history
        self.times = np.empty((1, 0), dtype=np.float64)
        self.trajectory = np.empty((0, 15), dtype=np.float64)

        # set integration method
        self._integrator.set_integrator("dop853", atol=atol, rtol=rtol)

        # set recorder
        self._integrator.set_solout(self._recorder)

        # set initial conditions
        self._integrator.set_initial_value(fs0, t0)

        # numerically integrate
        self._integrator.integrate(tf)

    def mismatch_constraints(self, atol=1e-5, rtol=1e-5):
        """Returns the nondimensional mismatch equality constraints of the arrival boundary conditions.

        This method propagates the nondimensional dynamics of the spacecraft
        from the departure time `t0` to the arrival time `tf`, then evaluates
        the nondimensional mismatch between the desired arrival state `xf`,
        arrival mass costate `lmf == 0` (`if freemass == True`), and arrival
        Hamiltonian `H == 0` (`if freetime == True`) and the supplied desired
        arrival boundary conditions.

        Args:
            - atol (``float``, ``int``): Absolute integration solution tolerance.
            - rtol (``float``, ``int``): Relative integration solution tolerance.

        Returns:
            - ``numpy.ndarray``: Equality constraints vector ``ceq`` composed of the arrival mismatch in position & velocity, arrival mass costate if ``freemass == True``, arrival mismatch in mass if ``freemass == False``, and arrival Hamiltonian if ``freetime == True``.

            ::

                ceq = [drf, dvf, dmf]    # freemass == False; freetime == False
                ceq = [drf, dvf, lmf]    # freemass == True; freetime == False
                ceq = [drf, dvf, lmf, H] # freemass == True; freetime == True

        Raises:
            - TypeError: If either ``atol`` or ``rtol`` is supplied as neither an instance of ``float`` nor ``int``.
            - AttributeError: If boundary conditions ``t0``, ``x0``, ``l0``, ``tf``, and ``x0`` have not been set through either ``__init__`` or ``set``.

        .. note::
            This method uses an explicit Runge-Kutta method of order (4)5
            due to Dormand & Prince with adaptive stepsize control.
            Smaller values of ``atol`` and ``rtol`` will increase the accuracy of
            a converged trajectory optimisation solution. However, smaller
            values result in slower integration executions and thus slower
            optimiser iterations. It is preferable to first solve a trajectory
            optimisation problem with higher values of ``atol`` and ``rtol`` to
            quickly converge to an approximate solution, subsequently resolving
            the problem with smaller tolerance values using the previous solution.
            This method is akin to mesh refinement in direct trajectory optimisation.

        .. note::
            State arrival boundary conditions in position and velocity are
            always returned. If the final mass transversality condition is
            activated (``freemass == True``), than the final mass costate is
            returned as part of the equality constraint vector. Otherwise
            (``freemass == False``), the final mass mismatch is returned as part
            of the equality constraint vector. If the final time transversality
            condition is activated ``freetime == True``, then than the final
            Hamiltonian will be returned as part of the equality constraint
            vector. Otherwise (``freetime == False``), the Hamiltonian will not
            be supplied as part of the equality constraint vector.

        Examples:
            >>> l = pykep.pontryagin.leg(freemass=True, freetime=True)
            >>> l.set(t0, x0, l0, tf, xf)
            >>> l.mismatch_constraints(atol=1e-10, rtol=1e-10)
            [ -1.15402379e+01  -2.23345886e+00  -9.30022917e-01  -2.53440778e+00
              -3.44246359e+00  -3.96669697e-01  -2.82967399e+03  -8.58526037e-01]
            >>> l.nec
            8

            >>> l = pykep.pontryagin.leg(freemass=False, freetime=True)
            >>> l.set(t0, x0, l0, tf, xf)
            >>> l.mismatch_constraints(atol=1e-10, rtol=1e-10)
            [-0.82435194  0.4375439   0.04615264  0.69192818 -0.18327442 -0.00930848
              0.46335841  1.66752446]
            >>> l.nec
            8

            >>> l = pykep.pontryagin.leg(freemass=True, freetime=False)
            >>> l.set(t0, x0, l0, tf, xf)
            >>> l.mismatch_constraints(atol=1e-10, rtol=1e-10)
            [  4.95782514e+00  -7.94403974e+00  -1.10158930e-02   7.27278513e+00
              -1.70998019e+00   9.13925064e-02  -1.45194673e+06]
            >>> l.nec
            7

            >>> l = pykep.pontryagin.leg(freemass=False, freetime=False)
            >>> l.set(t0, x0, l0, tf, xf)
            >>> l.mismatch_constraints(atol=1e-10, rtol=1e-10)
            [-0.90868565  0.23238016  0.04596579  0.61543688 -0.50023407 -0.0058185
              0.62170522]
            >>> l.nec
            7

            >>> l = pykep.pontryagin.leg()
            >>> l.mismatch_constraints(atol=1e-10, rtol=1e-10)
            AttributeError: Cannot propagate dynamics, as boundary conditions t0, x0, l0, tf, and xf have not been set. Use set(t0, x0, l0, tf, xf) to set boundary conditions.
        """

        if not all([(isinstance(tol, float) or isinstance(tol, int)) for tol in [atol, rtol]]):
            raise TypeError(
                "Both atol and rtol must be supplied as instances of either float or int.")
        if any([hasattr(self, atr) == False for atr in ["t0", "x0", "l0", "tf", "xf"]]):
            raise AttributeError(
                "Cannot propagate dynamics, as boundary conditions t0, x0, l0, tf, and xf have not been set. Use set(t0, x0, l0, tf, xf) to set boundary conditions.")
        else:
            atol = float(atol)
            rtol = float(atol)

        # propagate trajectory
        self._propagate(atol, rtol)

        # desired nondimensional arrival states
        brf = self.xf[0:3] / self._dynamics.L
        bvf = self.xf[3:6] / self._dynamics.V

        # propagated nondimensional arrival states
        rf = self.trajectory[-1, 0:3]
        vf = self.trajectory[-1, 3:6]

        # nondimensional position and velocity arrival mismatch
        drf = rf - brf
        dvf = (vf - bvf)

        # free arrival mass
        if self.freemass:
            lmf = self.trajectory[-1, 13]
        # fixed arrival mass
        else:
            dmf = self.trajectory[-1, 6] - self.xf[6] / self._dynamics.M

        # free arrival time
        if self.freetime:
            Hf = self._dynamics._hamiltonian(self.trajectory[-1])
        # fixed arrival time
        else:
            pass

        # create equality constraints
        if (self.freemass and self.freetime):
            ceq = np.hstack((drf, dvf, [lmf], [Hf]))
        elif (self.freemass and not self.freetime):
            ceq = np.hstack((drf, dvf, [lmf]))
        elif (self.freetime and not self.freemass):
            ceq = np.hstack((drf, dvf, [dmf], [Hf]))
        elif (not self.freemass and not self.freetime):
            ceq = np.hstack((drf, dvf, [dmf]))
        else:
            raise AttributeError(
                "Could not determine equality constraint vector.")

        return ceq

    def get_states(self, atol=1e-12, rtol=1e-12):
        """Returns the trajectory data of time, states, costates, and controls.

        This method propagates the spacecrafts dynamics with a chosen ``atol``
        ``rtol`` from ``t0``, ``x0``, ``l0`` to ``t0``, then returns a `numpy.ndarray`
        with the number of rows corresponding to the number of data points
        along the trajectory (as determinded by ``atol`` and ``rtol``), and 19
        columns.

        Args:
            - atol (``float``, ``int``): Absolute integration solution tolerance.
            - rtol (``float``, ``int``): Relative integration solution tolerance.

        Returns:
            - ``numpy.ndarray``: Trajectory data array of shape ``(npts, 20)``, where
              ``npts`` is the number of data points along the trajectory, as determind by the integration
              error tolerances ``atol`` and ``rtol``.

            The returned array is characterised by:
            ::

                [[t0, x0, y0, z0, vx0, vy0, vz0, m0, lx0, ly0, lz0, lvx0, lvy0, lvz0, lm0, obj, u0, ux0, uy0, uz0, H0],
                ...
                 [tf, xf, yf, zf, vxf, vyf, vzf, mf, lxf, lyf, lzf, lvxf, lvyf, lvzf, lmf, obj, uf, uxf, uyf, uzf, Hf]]

        Raises:
            - TypeError: If either ``atol`` or ``rtol`` is supplied as neither an instance of ``float`` nor ``int``.
            - AttributeError: If boundary conditions ``t0``, ``x0``, ``l0``, ``tf``, and ``x0`` have not been set through either ``__init__`` or ``set``.

        .. note::
            The returned array has the units of
            [mjd2000, m, m, m, m/s, m/s, m/s, kg, ND, ND, ND, ND, ND, ND, ND, s, ND, ND, ND, ND, ND].

        .. note::
            Setting either ``atol`` or ``rtol`` smaller than ``1e-12`` may result in numerical integration difficulties.

        Examples:
            >>> l = leg(t0, x0, l0, tf, xf)
            >>> traj = l.get_states(atol=1e-12, rtol=1e-12)
            >>> t = traj[:, 0] # times
            >>> r = traj[:, 1:4] # positions
            >>> v = traj[:, 4:7] # velocities
            >>> m = traj[:, 7] # masses
            >>> lm = traj[:, 14] # mass costates
            >>> u = traj[:, 15] # control throttles
            >>> H = traj[:, 19] # Hamiltonians
            >>> plt.plot(t, u) # plot the control throttle history
            >>> plt.show()
        """

        if not all([(isinstance(tol, float) or isinstance(tol, int)) for tol in [atol, rtol]]):
            raise TypeError(
                "Both atol and rtol must be supplied as instances of either float or int.")
        if any([hasattr(self, atr) == False for atr in ["t0", "x0", "l0", "tf", "xf"]]):
            raise AttributeError(
                "Cannot propagate dynamics, as boundary conditions t0, x0, l0, tf, and xf have not been set. Use set(t0, x0, l0, tf, xf) to set boundary conditions.")
        else:
            atol = float(atol)
            rtol = float(atol)

        # propagate trajectory
        self._propagate(atol, rtol)

        # nondimensional times
        t = np.copy(self.times)
        # mjs2000 times
        t *= self._dynamics.T
        # mjd2000 times
        t /= 24 * 60 * 60
        # reshape
        t = t.reshape(t.size, 1)

        # controls
        u = np.asarray([self._dynamics._pontryagin(fs)
                        for fs in self.trajectory])

        # get Hamiltonian
        H = np.asarray([self._dynamics._hamiltonian(fs)
                        for fs in self.trajectory])
        H = H.reshape(H.size, 1)

        # get trajectory
        traj = np.copy(self.trajectory)
        # redimensionalise trajectory
        traj[:, 0:3] *= self._dynamics.L
        traj[:, 3:6] *= self._dynamics.V
        traj[:,   6] *= self._dynamics.M

        # assemble full trajectory history
        traj = np.hstack((t, traj, u, H))

        return traj

    def plot_traj(self, axes, mark="k", atol=1e-11, rtol=1e-11, units=AU, quiver=False, length=1):
        """Plots trajectory onto a 3D axis.

        Args:
            - axes (``matplotlib.axes._subplots.Axes3DSubplot``): 3D axis onto which to plot the trajectory.
            - mark (``str``): Marker style.
            - atol (``float``, ``int``): Absolute integration solution tolerance.
            - rtol (``float``, ``int``): Relative integration solution tolerance.
            - units (``float``, ``int``): Length unit by which to normalise data.
            - quiver (``bool``): Activates the visualization of the throttle arrows
            - length (``float``): Length of thrust arrow if quiver is True

        Raises:
            - TypeError: If ``axes`` is not an instance of ``mpl_toolkits.mplot3d.Axes3D``.
            - TypeError: If ``mark`` is not an instance of ``str``.
            - TypeError: If either ``atol``, ``rtol``, or ``units`` is neither an instance of ``float`` nor ``int``.

        Examples:
            >>> sc = pk.sims_flanagan.spacecraft(1000, 0.3, 2500) # spacecraft
            >>> p0 = pk.planet.jpl_lp("earth")
            >>> pf = pk.planet.jpl_lp("mars")
            >>> t0 = pk.epoch(0)
            >>> tf = pk.epoch(1000)
            >>> r0, v0 = p0.eph(t0)
            >>> rf, vf = pf.eph(tf)
            >>> x0 = pk.sims_flanagan.sc_state(r0, v0, sc.mass)
            >>> xf = pk.sims_flanagan.sc_state(rf, vf, sc.mass/10)
            >>> l0 = np.random.randn(7)
            >>> l = leg(t0, x0, l0, tf, xf)
            >>> fig = plt.figure()
            >>> axes = fig.gca(projection='3d')
            >>> l.plot(axes)
            >>> plt.show()
        """

        if not isinstance(axes, Axes3D):
            raise TypeError(
                "Axis must be instance of matplotlib.axes._subplots.Axes3DSubplot.")
        elif not isinstance(mark, str):
            raise TypeError("Mark must be instance of string.")
        elif not all((isinstance(par, float) or isinstance(par, int)) for par in [atol, rtol, units]):
            raise TypeError(
                "atol, rtol, and units must be either instances of float or int.")
        else:
            pass

        # get trajectory
        full_data = self.get_states(atol=atol, rtol=rtol)
        traj = full_data[:, 1:4]

        # normalise
        traj /= units

        # plot trajectory
        axes.plot(traj[:, 0], traj[:, 1], traj[:, 2], mark)

        if quiver:
            thrusts = full_data[:, 17:20]
            magnitudes = full_data[:, 16]
            thrusts[:, 0] = thrusts[:, 0] * magnitudes
            thrusts[:, 1] = thrusts[:, 1] * magnitudes
            thrusts[:, 2] = thrusts[:, 2] * magnitudes

            axes.set_aspect("equal")
            xlim = axes.get_xlim()
            zlim = axes.get_zlim()
            ratio = (xlim[1] - xlim[0]) / (zlim[1] - zlim[0])
            # to compensate for non equal z axes
            thrusts[:, 2] = thrusts[:, 2] / ratio

            axes.quiver(traj[:, 0], traj[:, 1], traj[:, 2], thrusts[:, 0], thrusts[:, 1], thrusts[
                        :, 2], color='r', length=length, normalize=False, arrow_length_ratio=0.00, alpha=0.6)

        return axes

    def plot(self, x, y, mark="k.-", atol=1e-12, rtol=1e-12, unitsx=1, unitsy=1, xlabel=False, ylabel=False, axes=None):
        """Plots in two dimensions of the leg's trajectory data.
        ::

            keys = ['t', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'm', 'lx', 'ly', 'lz', 'lvx', 'lvy', 'lvz', 'lm', 'u', 'ux', 'uy', 'uz', 'H']

        Args:
            - x (``str``): x-axis dimension in ``keys``.
            - y (``str``): y-axis dimension in ``keys``.
            - mark (``str``): Marker style.
            - atol (``float``, ``int``): Absolute integration solution tolerance.
            - rtol (``float``, ``int``): Relative integration solution tolerance.
            - unitsx (``float``, ``int``): Unit by which to normalise x-axis data.
            - unitsy (``float``, ``int``): Unit by which to normalise y-axis data.
            - xlabel (``str``, ``bool``): x-axis label. If label is ``False``, no label is placed; if ``True``, dimension name is placed.
            - ylabel (``str``, ``bool``): y-axis label. If label is ``False``, no label is placed; if ``True``, dimension name is placed.

        """

        keys = ['tof', 't', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'm', 'lx', 'ly',
                'lz', 'lvx', 'lvy', 'lvz', 'lm', 'obj', 'u', 'ux', 'uy', 'uz', 'H']

        if not all([isinstance(dim, str) for dim in [x, y, mark]]):
            raise TypeError(
                "x, y, and mark must be supplied as instances of str.")
        elif not all([dim in keys for dim in [x, y]]):
            raise ValueError("Both x and y must be in " + str(keys) + ".")
        elif not all([(isinstance(par, float) or isinstance(par, int)) for par in [atol, rtol, unitsx, unitsy]]):
            raise TypeError(
                "atol, rtol, unitsx, and unitsy must be supplied as an instance of either int or float.")
        elif not all([(isinstance(label, str) or isinstance(label, bool)) for label in [xlabel, ylabel]]):
            raise TypeError(
                "xlabel and ylabel must be supplied an instance of either str or bool.")
        else:

            # get trajectory
            traj = self.get_states(atol=atol, rtol=rtol)

            # append tof
            tof = traj[:, 0]
            tof = tof - tof[0]
            traj = np.hstack((tof.reshape(tof.size, 1), traj))
            # get components and normalise
            xi, yi = keys.index(x), keys.index(y)

            # create figure
            if axes is None:
                fig = plt.figure()
                axes = fig.gca()

            # plot
            axes.plot(traj[:, xi] / unitsx, traj[:, yi] / unitsy, mark)

            # labels
            if isinstance(xlabel, str):
                plt.xlabel(xlabel)
            elif xlabel is True:
                plt.xlabel(x)

            if isinstance(ylabel, str):
                plt.ylabel(ylabel)
            elif ylabel is True:
                plt.ylabel(y)

            return axes



from pykep.trajopt import mga_1dsm
from pykep.planet import jpl_lp, keplerian
from pykep import AU, DEG2RAD, MU_SUN, epoch

# ROSETTA (we need to modify the safe radius of the planets to match the wanted problem)
_churyumov = keplerian(epoch(52504.23754000012, "mjd"),
                                 [3.50294972836275 * AU,
                                  0.6319356,
                                  7.12723 * DEG2RAD,
                                  50.92302 * DEG2RAD,
                                  11.36788 * DEG2RAD,
                                  0. * DEG2RAD],
                                 MU_SUN, 0., 0., 0., "Churyumov-Gerasimenko")

_mars_rosetta = jpl_lp('mars')
_mars_rosetta.safe_radius = 1.05
_seq_rosetta = [jpl_lp('earth'), 
                jpl_lp('earth'), 
                _mars_rosetta, 
                jpl_lp('earth'), 
                jpl_lp('earth'), 
                _churyumov]

class _rosetta_udp(mga_1dsm):
    """
    This class represents a rendezvous mission to the comet 67P/Churyumov-Gerasimenko modelled as an MGA-1DSM transfer.
    The fly-by sequence selected (i.e. E-EMEE-C) is similar to the one planned for the spacecraft Rosetta. 
    The objective function considered is the total mission delta V. No launcher model is employed and a final randezvous
    with the comet is included in the delta V computations.

    .. note::

       A significantly similar version of this problem was part of the no longer maintained GTOP database, 
       https://www.esa.int/gsp/ACT/projects/gtop/gtop.html. The exact definition is, though, different and results
       should thus not be compared to those posted in GTOP.
    """
    def __init__(self):
        """
        Write Me
        """
        super().__init__(
            seq =_seq_rosetta,
            t0 = [1460, 1825],
            tof = [[300, 500], [150, 800], [150, 800], [300, 800], [700, 1850]],
            vinf = [3., 5.],
            add_vinf_dep = False,
            add_vinf_arr = True,
            tof_encoding = "direct",
            multi_objective = False,
            eta_lb = 0.01,
            eta_ub = 0.9,
            rp_ub = 9.
        )

    def get_name(self):
        return "Rosetta (Trajectory Optimisation Gym P10)"

    def __repr__(self):
        return self.get_name()

# Problem P10: Rosetta mission MGA1DSM, single objective, direct encoding
rosetta = _rosetta_udp()


# -*- coding: iso-8859-1 -*-
# We import the proteced symbols we use in this file
from .core import _get_AU, _get_JR, _get_DAY2SEC, _get_DAY2YEAR, _get_DEG2RAD, _get_EARTH_VELOCITY, _get_EARTH_J2, _get_EARTH_RADIUS, _get_MU_EARTH, _get_G0, _get_MU_SUN, _get_RAD2DEG, _get_SEC2DAY
from .core import _epoch_type
# We import symbols we use in this file
from .core import epoch
# We import all symbols in the core namespace (also the ones we do not use
# in this file, but we still want in the namespace core)
from .core import *

"""Defining astronomical constants defined in the keplerian_toolbox file astro_constants.h"""
AU = _get_AU()
JR = _get_JR()
DAY2SEC = _get_DAY2SEC()
DAY2YEAR = _get_DAY2YEAR()
DEG2RAD = _get_DEG2RAD()
EARTH_VELOCITY = _get_EARTH_VELOCITY()
EARTH_J2 = _get_EARTH_J2()
EARTH_RADIUS = _get_EARTH_RADIUS()
G0 = _get_G0()
MU_SUN = _get_MU_SUN()
MU_EARTH = _get_MU_EARTH()
RAD2DEG = _get_RAD2DEG()
SEC2DAY = _get_SEC2DAY()

EPOCH_TYPE = {
    "jd": _epoch_type.JD,
    "mjd": _epoch_type.MJD,
    "mjd2000": _epoch_type.MJD2000
}


def _epoch_ctor(self, julian_date=0, julian_date_type='mjd2000'):
    """
pykep.epoch(julian_date=0, julian_date_type="mjd2000")

- julian_date: a julian date, defaults to 0
- julian_date_type: julian date type, one of "jd", "mjd" and "mjd2000", defaults to "mjd2000"

Examples::

  e1 = epoch(0)
  e1 = epoch(0,"mjd2000")
  e2 = epoch(54333, "mjd")
    """
    self._orig_init(julian_date, EPOCH_TYPE[julian_date_type])

epoch._orig_init = epoch.__init__
epoch.__init__ = _epoch_ctor



class dbscan():
    """
    This class can be used to locate areas of the interplanetsry space that are 'dense' at one epoch.
    Essentially, it locates planet clusters
    """
    from pykep.core import AU, EARTH_VELOCITY

    def _axis_equal_3d(self, ax):
        """Rescales 3D axis limits using equal scale."""
        import numpy
        extents = numpy.array(
            [getattr(ax, 'get_{}lim'.format(dim))() for dim in 'xyz'])
        sz = extents[:, 1] - extents[:, 0]
        centers = numpy.mean(extents, axis=1)
        maxsize = max(abs(sz))
        r = maxsize / 2
        for ctr, dim in zip(centers, 'xyz'):
            getattr(ax, 'set_{}lim'.format(dim))(ctr - r, ctr + r)

    def __init__(self, planet_list):
        """
        USAGE: cl = dbscan(planet_list):

        - planet_list = list of pykep planets (typically thousands)
        """
        self._asteroids = planet_list
        self.labels = None
        self.n_clusters = None
        self.members = None
        self.core_members = None

    def _orbital_metric(self, r, v):
        from pykep.core import DAY2SEC
        DV2 = [a / (self._T * DAY2SEC) for a in r]
        DV1 = [a + b for a, b in zip(DV2, v)]
        return DV1 + DV2

    def cluster(self, t, eps=0.125, min_samples=10, metric='orbital', T=180, ref_r=AU, ref_v=EARTH_VELOCITY):
        """
        USAGE: cl.cluster(t, eps=0.125, min_samples=10, metric='orbital', T=180, ref_r=AU, ref_v=EARTH_VELOCITY):

        - t: epoch (in MJD2000)
        - eps: max distance between points in a cluster
        - min_samples: minimum number of samples per cluster
        - metric: one of 'euclidean', 'euclidean_r', orbital'
        - T: average transfer time (used in the definition of the 'orbital' metric)
        - ref_r         reference radius   (used as a scaling factor for r if the metric is 'euclidean' or 'euclidean_r')
        - ref_v         reference velocity (used as a scaling factor for v if the metric is 'euclidean')
        """
        import pykep
        import numpy
        from sklearn.cluster import DBSCAN

        self._epoch = pykep.epoch(t)

        if metric == 'euclidean':
            self._X = [
                [elem for tupl in p.eph(self._epoch) for elem in tupl] for p in self._asteroids]
            scaling_vector = [ref_r] * 3
            scaling_vector += [ref_v] * 3
        elif metric == 'euclidean_r':
            self._X = [list(p.eph(self._epoch)[0]) for p in self._asteroids]
            scaling_vector = [ref_r] * 3
        elif metric == 'orbital':
            self._T = T
            self._X = [self._orbital_metric(
                *p.eph(self._epoch)) for p in self._asteroids]
            scaling_vector = [1.] * 6  # no scaling
        self._X = numpy.array(self._X)

        scaling_vector = numpy.array(scaling_vector)
        self._X = self._X / scaling_vector[None, :]

        self._db = DBSCAN(eps=eps, min_samples=min_samples).fit(self._X)
        self._core_samples = self._db.core_sample_indices_

        self.labels = self._db.labels_
        self.n_clusters = len(
            set(self.labels)) - (1 if -1 in self.labels else 0)

        self.members = {}
        self.core_members = {}
        for label in set(self.labels):
            if int(label) == -1:
                continue
            self.members[int(label)] = [index[0]
                                        for index in numpy.argwhere(self.labels == label)]
            self.core_members[int(label)] = [
                index for index in self._core_samples if self.labels[index] == label]

        self._X = self._X * scaling_vector[None, :]

    def pretty(self):
        """Prints the cluster lists."""
        if self.labels is None:
            return
        print("Number of clusters: %d" % self.n_clusters)
        print("Size of dataset: %s" % str(self._X.shape))
        print("Scaling: %s" % str(self._scaling))
        print("Epoch: %s" % str(self._epoch))
        for label in list(self.members.keys()):
            print("cluster %d (%d - %d): %s" % (label, len(self.members[label]),
                                                len(self.core_members[label]), str(self.members[label])))

    def plot(self, ax=None, clusters=None, orbits=False, only_core=False):
        """Plots the clusters."""
        if self.n_clusters < 1:
            return

        import matplotlib.pylab as plt
        from mpl_toolkits.mplot3d import Axes3D

        if ax is None:
            fig = plt.figure()
            axis = fig.add_subplot(111, projection='3d')
        else:
            axis = ax

        axis.view_init(elev=30.0, azim=135.0)
        axis.set_aspect('equal')

        if orbits:
            from pykep.orbit_plots import plot_planet
            members = self.core_members if only_core else self.members
            for label in members if clusters is None else clusters:
                for planet in members[label]:
                    plot_planet(
                        self._asteroids[planet], t0=self._epoch, s=0, axes=axis)

        X, labels = list(zip(*[(x, label) for (x, label) in zip(self._X, self.labels)
                               if label > -.5 and (clusters is None or label in clusters)]))
        data = [[x[0], x[1], x[2]] for x in X]
        axis.scatter(*list(zip(*data)), c=labels, alpha=0.5)

        self._axis_equal_3d(axis)

        if ax is None:
            plt.show()
        return axis

    def plot_cluster_evolution(self, cluster_id=None, only_core=False, epochs=range(7500, 8400, 100), skip=100, alpha=0.3):
        """
        Plots a cluster evolution at 9 prefixed epochs.


        """
        if self.n_clusters < 1:
            print("No clusters have been found yet")
            return
        if cluster_id >= self.n_clusters or cluster_id < 0:
            print(
                "cluster_id should be larger then 0 and smaller than the number of clusters (-1)")
            return
        if len(epochs) != 9:
            print("The epochs requested must be exactly 9 as to assemble 3x3 subplots")
            return

        import matplotlib.pylab as plt
        from mpl_toolkits.mplot3d import Axes3D
        from pykep.orbit_plots import plot_planet
        from pykep import epoch

        if only_core:
            ids = self.core_members[cluster_id]
        else:
            ids = self.members[cluster_id]

        fig = plt.figure()
        for i, ep in enumerate(epochs):
            axis = fig.add_subplot(3, 3, i + 1, projection='3d')

            plt.axis('off')
            plt.title(epoch(ep).__repr__()[:11])
            for pl in self._asteroids[::skip]:
                axis = plot_planet(pl, axes=axis, alpha=0.05, s=0)
            for cluster_member in ids:
                r, _ = self._asteroids[cluster_member].eph(epoch(ep))
                axis.scatter([r[0]], [r[1]], [r[2]], marker='o', alpha=alpha)

        plt.draw()
        plt.show()
        return fig



import pykep as pk


class mga_lt_EVMe(object):

    """
    This constructs a pagmo udp that represents a low-thrust transfer between Earth and Mercury with a Venus
    fly-by. The decision vector contains
    [
    t0,
    T1, mf1, Vxi1, Vyi1, Vzi1, Vxf1, Vyf1, Vzf1,
    T2, mf2, Vxi2, Vyi2, Vzi2, Vxf2, Vyf2, Vzf2,
    [throttles1], [throttles2]
    ]
    in the following units: [mjd2000, days, kg, m/s,m/s,m/s, [non-dimensional]]
    """

    def __init__(self, mass=2000, Tmax=0.5, Isp=3500, Vinf_dep=3, Vinf_arr=2, nseg1=5, nseg2=20):
        # We define some data members (we use the double underscore to
        # indicate they are private)
        self.__earth = pk.planet.jpl_lp('earth')
        self.__venus = pk.planet.jpl_lp('venus')
        self.__mercury = pk.planet.jpl_lp('mercury')
        self.__sc = pk.sims_flanagan.spacecraft(mass, Tmax, Isp)
        self.__Vinf_dep = Vinf_dep * 1000
        self.__Vinf_arr = Vinf_arr * 1000
        self.__leg1 = pk.sims_flanagan.leg()
        self.__leg2 = pk.sims_flanagan.leg()
        self.__leg1.set_mu(pk.MU_SUN)
        self.__leg1.set_spacecraft(self.__sc)
        self.__leg2.set_mu(pk.MU_SUN)
        self.__leg2.set_spacecraft(self.__sc)
        self.__nseg1 = nseg1
        self.__nseg2 = nseg2
        self.__mass = mass

    def get_nec(self):
        return 15

    def get_nic(self):
        return self.__nseg1 + self.__nseg2 + 3

    def get_bounds(self):
        lb = [3000, 100, self.__mass / 2] + [-self.__Vinf_dep] * 3 + [-6000] * 3 + [200, self.__mass /
                                                                                    9] + [-6000] * 3 + [-self.__Vinf_arr] * 3 + [-1, -1, -1] * (self.__nseg1 + self.__nseg2)
        ub = [4000, 1000, self.__mass] + [self.__Vinf_dep] * 3 + [6000] * 3 + [2000, self.__mass] + \
            [6000] * 3 + [self.__Vinf_arr] * 3 + \
            [1, 1, 1] * (self.__nseg1 + self.__nseg2)
        return (lb, ub)

    # This is the objective function
    def fitness(self, x):
        from pykep import epoch, AU, EARTH_VELOCITY, fb_con
        from pykep.sims_flanagan import leg, sc_state
        from numpy.linalg import norm
        from math import sqrt, asin, acos

        retval = [-x[10]]

        # Ephemerides
        t_E = epoch(x[0])
        t_V = epoch(x[0] + x[1])
        t_M = epoch(x[0] + x[1] + x[9])
        rE, vE = self.__earth.eph(t_E)
        rV, vV = self.__venus.eph(t_V)
        rM, vM = self.__mercury.eph(t_M)

        # First Leg
        v = [a + b for a, b in zip(vE, x[3:6])]
        x0 = sc_state(rE, v, self.__sc.mass)
        v = [a + b for a, b in zip(vV, x[6:9])]
        xe = sc_state(rV, v, x[2])
        self.__leg1.set(
            t_E, x0, x[-3 * (self.__nseg1 + self.__nseg2):-self.__nseg2 * 3], t_V, xe)

        # Second leg
        v = [a + b for a, b in zip(vV, x[11:14])]
        x0 = sc_state(rV, v, x[2])
        v = [a + b for a, b in zip(vM, x[14:17])]
        xe = sc_state(rM, v, x[10])
        self.__leg2.set(t_E, x0, x[(-3 * self.__nseg2):], t_V, xe)

        # Defining the constraints
        # departure
        v_dep_con = (x[3] * x[3] + x[4] * x[4] + x[5] * x[5] -
                     self.__Vinf_dep * self.__Vinf_dep) / (EARTH_VELOCITY * EARTH_VELOCITY)
        # arrival
        v_arr_con = (x[14] * x[14] + x[15] * x[15] + x[16] * x[16] -
                     self.__Vinf_arr * self.__Vinf_arr) / (EARTH_VELOCITY * EARTH_VELOCITY)
        # fly-by at Venus
        DV_eq, alpha_ineq = fb_con(x[6:9], x[11:14], self.__venus)

        # Assembling the constraints
        constraints = list(self.__leg1.mismatch_constraints() + self.__leg2.mismatch_constraints()) + [DV_eq] + list(
            self.__leg1.throttles_constraints() + self.__leg2.throttles_constraints()) + [v_dep_con] + [v_arr_con] + [alpha_ineq]

        # We then scale all constraints to non-dimensional values
        # leg 1
        constraints[0] /= AU
        constraints[1] /= AU
        constraints[2] /= AU
        constraints[3] /= EARTH_VELOCITY
        constraints[4] /= EARTH_VELOCITY
        constraints[5] /= EARTH_VELOCITY
        constraints[6] /= self.__sc.mass
        # leg 2
        constraints[7] /= AU
        constraints[8] /= AU
        constraints[9] /= AU
        constraints[10] /= EARTH_VELOCITY
        constraints[11] /= EARTH_VELOCITY
        constraints[12] /= EARTH_VELOCITY
        constraints[13] /= self.__sc.mass
        # fly-by at Venus
        constraints[14] /= (EARTH_VELOCITY * EARTH_VELOCITY)

        return retval + constraints

    # And this helps to visualize the trajectory
    def plot(self, x):
        import matplotlib as mpl
        from mpl_toolkits.mplot3d import Axes3D
        import matplotlib.pyplot as plt
        from pykep import epoch, AU
        from pykep.sims_flanagan import sc_state
        from pykep.orbit_plots import plot_planet, plot_sf_leg

        t_E = epoch(x[0])
        t_V = epoch(x[0] + x[1])
        t_M = epoch(x[0] + x[1] + x[9])
        rE, vE = self.__earth.eph(t_E)
        rV, vV = self.__venus.eph(t_V)
        rM, vM = self.__mercury.eph(t_M)

        # First Leg
        v = [a + b for a, b in zip(vE, x[3:6])]
        x0 = sc_state(rE, v, self.__sc.mass)
        v = [a + b for a, b in zip(vV, x[6:9])]
        xe = sc_state(rV, v, x[2])
        self.__leg1.set(
            t_E, x0, x[-3 * (self.__nseg1 + self.__nseg2):-self.__nseg2 * 3], t_V, xe)

        # Second leg
        v = [a + b for a, b in zip(vV, x[11:14])]
        x0 = sc_state(rV, v, x[2])
        v = [a + b for a, b in zip(vM, x[14:17])]
        xe = sc_state(rM, v, x[10])
        self.__leg2.set(t_E, x0, x[(-3 * self.__nseg2):], t_V, xe)

        fig = plt.figure()
        axis = fig.gca(projection='3d')

        # The Sun
        axis.scatter([0], [0], [0], color='y')
        # The legs
        plot_sf_leg(self.__leg1, units=AU, N=10, axes=axis)
        plot_sf_leg(self.__leg2, units=AU, N=10, axes=axis)
        # The planets
        plot_planet(
            self.__earth, t_E, units=AU, legend=True, color=(0.7, 0.7, 1), axes=axis)
        plot_planet(
            self.__venus, t_V, units=AU, legend=True, color=(0.7, 0.7, 1), axes=axis)
        plot_planet(
            self.__mercury, t_M, units=AU, legend=True, color=(0.7, 0.7, 1), axes=axis)
        plt.show()

"""
This example constructs, using pygmo for optimization, an interplanetary low-thrust optimization
problem that can then be solved using one of the available pygmo solvers. The problem is a non-linear constrained
problem that uses the Sims-Flanagan transcription to model the low-thrust trajectory. pykep plotting capabilities
are also demonstrated via the plot method. The interplanetary mission modelled is an LT-MGA Earth-Venus-Mercury mission.

"""


def run_example3():
    import pygmo as pg
    from pykep.examples import add_gradient, algo_factory

    # problem
    udp = add_gradient(mga_lt_EVMe(), with_grad=False)
    prob = pg.problem(udp)
    prob.c_tol = [1e-5] * prob.get_nc()

    # algorithm
    uda = algo_factory("snopt7", False)
    uda2 = pg.mbh(uda, 5, 0.05)
    algo = pg.algorithm(uda2)
    algo.set_verbosity(1)

    # 3 - Population
    pop = pg.population(prob, 1)

    # 4 - Solve the problem (evolve)
    print("Running Monotonic Basin Hopping ....")
    pop = algo.evolve(pop)

    print("Is the solution found a feasible trajectory? " +
          str(prob.feasibility_x(pop.champion_x)))
    udp.udp_inner.plot(pop.champion_x)

if __name__ == "__main__":
    run_example3()



def run_example11(n_seg=30):
    """
    This example demonstrates the use of the class lt_margo developed for the internal ESA CDF study on the 
    interplanetary mission named MARGO. The class was used to produce the preliminary traget selection
    for the mission resulting in 88 selected possible targets 
    (http://www.esa.int/spaceinimages/Images/2017/11/Deep-space_CubeSat)
    (http://www.esa.int/gsp/ACT/mad/projects/lt_to_NEA.html)
    """
    import pykep as pk
    import pygmo as pg
    import numpy as np
    from matplotlib import pyplot as plt
    from pykep.examples import add_gradient, algo_factory

    # 0 - Target asteroid from MPCORB
    mpcorbline = "K14Y00D 24.3   0.15 K1794 105.44160   34.12337  117.64264    1.73560  0.0865962  0.88781021   1.0721510  2 MPO369254   104   1  194 days 0.24 M-v 3Eh MPCALB     2803          2014 YD            20150618"

    # 1 - Algorithm
    algo = algo_factory("snopt7")

    # 2 - Problem
    udp = add_gradient(pk.trajopt.lt_margo(
        target=pk.planet.mpcorb(mpcorbline),
        n_seg=n_seg,
        grid_type="uniform",
        t0=[pk.epoch(8000), pk.epoch(9000)],
        tof=[200, 365.25 * 3],
        m0=20.0,
        Tmax=0.0017,
        Isp=3000.0,
        earth_gravity=False,
        sep=True,
        start="earth"),
        with_grad=False
    )
    prob = pg.problem(udp)
    prob.c_tol = [1e-5] * prob.get_nc()

    # 3 - Population
    pop = pg.population(prob, 1)

    # 4 - Solve the problem (evolve)
    pop = algo.evolve(pop)

    # 5 - Inspect the solution
    print("Feasible?:", prob.feasibility_x(pop.champion_x))

    # 6 - plot trajectory
    axis = udp.udp_inner.plot_traj(pop.champion_x, plot_segments=True)
    plt.title("The trajectory in the heliocentric frame")

    # 7 - plot control
    udp.udp_inner.plot_dists_thrust(pop.champion_x)

    # 8 - pretty
    udp.udp_inner.pretty(pop.champion_x)

    plt.ion()
    plt.show()


if __name__ == "__main__":
    run_example11()



from pykep.trajopt import mga
from pykep.planet import jpl_lp
from pykep import AU, DEG2RAD, MU_SUN, epoch

# CASSINI1
_seq_cassini1 = [jpl_lp('earth'), 
                jpl_lp('venus'), 
                jpl_lp('venus'), 
                jpl_lp('earth'), 
                jpl_lp('jupiter'), 
                jpl_lp('saturn')]

class _cassini1_udp(mga):
    """
    Write Me
    """
    def __init__(self):
        """
        Write Me
        """
        super().__init__(
            seq=_seq_cassini1,
            t0=[-1000., 0.],
            tof=[[30, 400], [100, 470], [30, 400], [400, 2000], [1000, 6000]],
            vinf=3.,
            tof_encoding='direct',
            orbit_insertion=True,
            e_target=0.98,
            rp_target=108950000)

    def get_name(self):
        return "Cassini MGA direct tof encoding (Trajectory Optimisation Gym P1)"

    def get_extra_info(self):
        retval = "\tTrajectory Optimisation Gym problem (P1): Cassini MGA, single objective, direct encoding\n"
        retval += "\tPlanetary sequence" + \
            str([pl.name for pl in _seq_cassini1])
        return retval

    def __repr__(self):
        return self.get_name()


class _cassini1a_udp(mga):
    """
    Write Me
    """
    def __init__(self):
        """
        Write Me
        """
        super().__init__(
            seq=_seq_cassini1,
            t0=[-1000., 0.],
            tof=[4000., 7000.],
            vinf=3.,
            tof_encoding='alpha',
            orbit_insertion=True,
            e_target=0.98,
            rp_target=108950000)

    def get_name(self):
        return "Cassini MGA alpha tof encoding (Trajectory Optimisation Gym P2)"

    def get_extra_info(self):
        retval = "\tTrajectory Optimisation Gym problem (P2): Cassini MGA, single objective, alpha encoding\n"
        retval += "\tPlanetary sequence" + \
            str([pl.name for pl in _seq_cassini1])
        return retval

    def __repr__(self):
        return self.get_name()


class _cassini1n_udp(mga):
    """
    Write Me
    """
    def __init__(self):
        """
        Write Me
        """
        super().__init__(
            seq=_seq_cassini1,
            t0=[-1000., 0.],
            tof=7000.,
            vinf=3.,
            tof_encoding='eta',
            orbit_insertion=True,
            e_target=0.98,
            rp_target=108950000)

    def get_name(self):
        return "Cassini1 MGA eta tof encoding (Trajectory Optimisation Gym P3)"

    def get_extra_info(self):
        retval = "\tTrajectory Optimisation Gym problem (P3): Cassini MGA, single objective, eta encoding\n"
        retval += "\tPlanetary sequence" + \
            str([pl.name for pl in _seq_cassini1])
        return retval

    def __repr__(self):
        return self.get_name()

# Problem P1: Cassini MGA, single objective, direct encoding
cassini1 = _cassini1_udp()
# Problem P2: Cassini MGA, single objective, alpha encoding
cassini1_a = _cassini1a_udp()
# Problem P3: Cassini MGA, single objective, eta encoding
cassini1_n = _cassini1n_udp()


def run_example6(n_seg=5):
    """
    This example demonstrates the optimization of a multiple randezvous mission (low-thrust).
    Such a mission (including more asteroids) is also called asteroid hopping

    The spacecraft performances, as well as the three asteroids visited, are taken from the GTOC7 problem description.
    """
    import pygmo as pg
    from pykep.trajopt import mr_lt_nep
    from pykep.planet import gtoc7
    from pykep.examples import add_gradient, algo_factory
    from matplotlib import pyplot as plt

    # If you have no access to snopt7, try slsqp (multiple starts may be
    # necessary)
    algo = algo_factory("snopt7")

    udp = add_gradient(
        mr_lt_nep(
            t0=[9600., 9700.],
            seq=[gtoc7(5318), gtoc7(14254), gtoc7(7422), gtoc7(5028)],
            n_seg=n_seg,
            mass=[800., 2000.],
            leg_tof=[100., 365.25],
            rest=[30., 365.25],
            Tmax=0.3,
            Isp=3000.,
            traj_tof=365.25 * 3.,
            objective='mass',
            c_tol=1e-05
        ),
        with_grad=False)

    prob = pg.problem(udp)
    prob.c_tol = [1e-5] * prob.get_nc()

    pop = pg.population(prob, 1)
    pop = algo.evolve(pop)

    solution = pop.champion_x
    if prob.feasibility_x(solution):
        print("FEASIBILE!!!")
        ax = udp.udp_inner.plot(solution)
    else:
        print("INFEASIBLE :(")
        ax = None

    plt.ion()
    plt.show()

if __name__ == "__main__":
    run_example6()



"""
This module contains functions that allow to study the planetary phasing.
That is the relative planetary position
"""
from pykep import __extensions__

if (__extensions__['scipy']):
    from ._knn import *

if (__extensions__['scikit-learn']):
    from ._dbscan import *

# if (__extensions__['pygmo'] and __extensions__['mplot3d']):
#    from ._lambert import lambert_metric


def three_impulses_approx(pl1, pl2, ep1=None, ep2=None):
    """
DV = pykep.phasing.three_impulses_approx(pl1, pl2, ep1=None, ep2=None)

- pl1: departure planet
- pl2: arrival planet
- ep1: departure epoch (optional and only useful for non keplerian planets)
- ep1: arrival epoch (default value is ep1).

- [out] DV: estimated DV cost for the orbital trab=nsfer

Returns the DV in m/s necessary for an orbit transfer between pl1 and pl2 assuming a perfect phasing.
The transfer will be made of three impulses. One to match apogees, one to match inclination and RAAN and
one to match perigees. Two of the three impulses will be merged together at either departure or arrival.
The argument of perigee is not matched, so that this approximation is only good for near-circular orbits.

Examples::

  DV = three_impulses_approx(pl1, pl2)
  DV = three_impulses_approx(pl1, pl2, ep1 = epoch(5500))
  DV = three_impulses_approx(pl1, pl2, ep1 = epoch(5500), ep2 = epoch(5700))
    """

    from pykep.core._core import _three_impulses_approx
    if ep2 is None:
        ep2 = ep1
    if ep1 is None:
        return _three_impulses_approx(pl1, pl2)
    return _three_impulses_approx(pl1, pl2, ep1, ep2)



import pykep as pk
import numpy as np
import math

mpcorbline = "K14Y00D 24.3   0.15 K1794 105.44160   34.12337  117.64264    1.73560  0.0865962  0.88781021   1.0721510  2 MPO369254   104   1  194 days 0.24 M-v 3Eh MPCALB     2803          2014 YD            20150618"


class lt_margo:
    """
    This class can be used as a User Defined Problem (UDP) in the pygmo2 software and if successfully solved,
    represents a low-thrust interplanetary trajectory from the Earth (or from the
    Sun-Earth L1 or L2 Lagrangian point) to a target NEO. The trajectory is modeled using the Sims-Flanagan model, 
    extended to include the Earth's gravity (assumed constant along each segment).
    The propulsion model can be both nuclear (NEP) or solar (SEP).

    This problem was developed and used during the European Space Agency interplanetary cubesat M-ARGO study.

    The decision vector (chromosome) is::

      [t0, tof, mf] + [throttles1] + [throttles2] + ...
    """

    def __init__(self,
                 target=pk.planet.mpcorb(mpcorbline),
                 n_seg=30,
                 grid_type="uniform",
                 t0=[pk.epoch(8000), pk.epoch(9000)],
                 tof=[200, 365.25 * 3],
                 m0=20.0,
                 Tmax=0.0017,
                 Isp=3000.0,
                 earth_gravity=False,
                 sep=False,
                 start="earth"):
        """
        prob = lt_margo(target = pk.planet.mpcorb(mpcorbline), n_seg = 30, grid_type = "uniform", t0 = [epoch(8000), epoch(9000)], tof = [200, 365.25*3], m0 = 20.0, Tmax = 0.0017, Isp = 3000.0, earth_gravity = False, sep = False, start = "earth")

        Args:
            - target (``pykep.planet``): target planet
            - n_seg (``int``): number of segments to use in the problem transcription (time grid)
            - grid_type (``string``): "uniform" for uniform segments, "nonuniform" toi use a denser grid in the first part of the trajectory
            - t0 (``list``): list of two pykep.epoch defining the bounds on the launch epoch
            - tof (``list``): list of two floats defining the bounds on the time of flight (days)
            - m0 (``float``): initial mass of the spacecraft (kg)
            - Tmax (``float``): maximum thrust at 1 AU (N)
            - Isp (``float``): engine specific impulse (s)
            - earth_gravity (``bool``): activates the Earth gravity effect in the dynamics
            - sep (``bool``): Activates a Solar Electric Propulsion model for the thrust - distance dependency.
            - start(``string``): Starting point ("earth", "l1", or "l2").

        .. note::

           L1 and L2 are approximated as the points on the line connecting the Sun and the Earth at a distance of, respectively, 0.99 and 1.01 AU from the Sun.

        .. note::

           If the Earth's gravity is enabled, the starting point cannot be the Earth
        """

        # 1) Various checks
        if start not in ["earth", "l1", "l2"]:
            raise ValueError("start must be either 'earth', 'l1' or 'l2'")
        if grid_type not in ["uniform", "nonuniform"]:
            raise ValueError(
                "grid_type must be either 'uniform' or 'nonuniform'")
        if earth_gravity and start == "earth":
            raise ValueError(
                "If Earth gravity is enabled the starting point cannot be the Earth")

        # 2) Class data members
        # public:
        self.target = target
        # private:
        self.__n_seg = n_seg
        self.__grid_type = grid_type
        self.__sc = pk.sims_flanagan._sims_flanagan.spacecraft(m0, Tmax, Isp)
        self.__earth = pk.planet.jpl_lp('earth')
        self.__earth_gravity = earth_gravity
        self.__sep = sep
        self.__start = start
        # grid construction
        if grid_type == "uniform":
            grid = np.array([i / n_seg for i in range(n_seg + 1)])
        elif grid_type == "nonuniform":
            grid_f = lambda x: x**2 if x < 0.5 else 0.25 + 1.5 * \
                (x - 0.5)  # quadratic in [0,0.5], linear in [0.5,1]
            grid = np.array([grid_f(i / n_seg) for i in range(n_seg + 1)])
        # index corresponding to the middle of the transfer
        fwd_seg = int(np.searchsorted(grid, 0.5, side='right'))
        bwd_seg = n_seg - fwd_seg
        fwd_grid = grid[:fwd_seg + 1]
        bwd_grid = grid[fwd_seg:]
        self.__fwd_seg = fwd_seg
        self.__fwd_grid = fwd_grid
        self.__fwd_dt = np.array([(fwd_grid[i + 1] - fwd_grid[i])
                                  for i in range(fwd_seg)]) * pk.DAY2SEC
        self.__bwd_seg = bwd_seg
        self.__bwd_grid = bwd_grid
        self.__bwd_dt = np.array([(bwd_grid[i + 1] - bwd_grid[i])
                                  for i in range(bwd_seg)]) * pk.DAY2SEC

        # 3) Bounds
        lb = [t0[0].mjd2000] + [tof[0]] + [0] + [-1, -1, -1] * n_seg
        ub = [t0[1].mjd2000] + [tof[1]] + [m0] + [1, 1, 1] * n_seg
        self.__lb = lb
        self.__ub = ub

    # Fitness function
    def fitness(self, x):
        obj = -x[2]

        ceq = list()
        cineq = list()
        throttles_constraints = []
        mismatch_constraints = []

        throttles = [x[3 + 3 * i: 6 + 3 * i] for i in range(self.__n_seg)]
        for t in throttles:
            throttles_constraints.append(t[0]**2 + t[1]**2 + t[2]**2 - 1.)
        cineq.extend(throttles_constraints)

        rfwd, rbwd, vfwd, vbwd, mfwd, mbwd, _, _, _, _, dfwd, dbwd = self._propagate(
            x)

        if self.__earth_gravity:
            flyby_constraints = []
            for d in dfwd + dbwd:
                d2 = d[0]**2 + d[1]**2 + d[2]**2
                l22 = (0.01 * pk.AU)**2
                flyby_constraints.append(1. - d2 / l22)
            cineq.extend(flyby_constraints)

        mismatch_constraints.extend([a - b for a, b in zip(rfwd[-1], rbwd[0])])
        mismatch_constraints.extend([a - b for a, b in zip(vfwd[-1], vbwd[0])])
        mismatch_constraints.append(mfwd[-1] - mbwd[0])
        ceq.extend(mismatch_constraints)

        # Making the mismatches non dimensional
        ceq[0] /= pk.AU
        ceq[1] /= pk.AU
        ceq[2] /= pk.AU
        ceq[3] /= pk.EARTH_VELOCITY
        ceq[4] /= pk.EARTH_VELOCITY
        ceq[5] /= pk.EARTH_VELOCITY
        ceq[6] /= self.__sc.mass

        # We assemble the constraint vector
        retval = [obj]
        retval.extend(ceq)
        retval.extend(cineq)

        return retval

    # Get bounds
    def get_bounds(self):
        return (self.__lb, self.__ub)

    # Get number of inequality contraints
    def get_nic(self):
        if self.__earth_gravity:
            return 2 * self.__n_seg
        else:
            return self.__n_seg

    # Get number of equality contraints
    def get_nec(self):
        return 7

    # Gradient sparsity
    def gradient_sparsity(self):
        # Objective
        retval = [[0, 2]]
        # Mismatches
        retval += [[i, j] for i in range(1, 7)
                   for j in range(3 * (1 + self.__n_seg))]
        retval += [[7, j] for j in range(1, 3 * (1 + self.__n_seg))]
        # Throttles
        retval += [[8 + i, 3 + 3 * i + j]
                   for i in range(self.__n_seg) for j in range(3)]
        # Prevent flyby
        if self.__earth_gravity:
            retval += [[8 + self.__n_seg + i, j]
                       for i in range(self.__fwd_seg) for j in [0, 1] + list(range(3, 6 + 3 * i))]
            retval += [[8 + self.__n_seg + self.__fwd_seg + i, j] for i in range(self.__bwd_seg) for j in [
                0, 1, 2] + list(range(3 + 3 * (self.__n_seg - self.__bwd_seg + i), 3 + 3 * self.__n_seg))]
        return retval

    def get_name(self):
        return "MARGO cubesat transfer to " + self.target.name

    def get_extra_info(self):
        retval = "\tTarget planet: " + self.target.name
        retval += "\n\tEarth gravity: " + str(self.__earth_gravity)
        retval += "\n\tSolar Electric Propulsion: " + str(self.__sep)
        retval += "\n\tStart mass: " + str(self.__sc.mass) + " kg"
        retval += "\n\tMaximum thrust as 1AU: " + str(self.__sc.thrust) + " N"
        retval += "\n\tSpecific impulse: " + str(self.__sc.isp) + " s"
        retval += "\n\n\tLaunch window: [" + \
            str(self.__lb[0]) + ", " + str(self.__ub[0]) + "] - MJD2000"
        retval += "\n\tBounds on time of flight: [" + str(
            self.__lb[1]) + ", " + str(self.__ub[1]) + "] - days"
        retval += "\n\n\tNumber of segments: " + str(self.__n_seg)
        retval += "\n\tGrid type: " + self.__grid_type

        return retval

    # SEP model
    def _sep_model(self, r):
        SAA = 0
        Psupply = 13.75
        eff = 0.92

        Pbmp = (-40.558 * r**3 + 173.49 * r**2 -
                259.19 * r + 141.86) * math.cos(SAA)
        P = -146.26 * r**3 + 658.52 * r**2 - 1059.2 * r + 648.24  # 6 panels
        # P = -195.02 * r**3 + 878.03 * r**2 - 1412.3 * r + 864.32 # 8 panels
        if Pbmp < Psupply:
            P -= (Psupply - Pbmp)
        Pin = eff * P
        if Pin > 120:
            Pin = 120  # thermal max 120W

        Tmax = (26.27127 * Pin - 708.973) / 1000000
        if Tmax < 0:
            Tmax = 0

        Isp = -0.0011 * Pin**3 + 0.175971 * Pin**2 + 4.193797 * Pin + 2037.213

        return Tmax, Isp

    # Propagates the trajectory
    def _propagate(self, x):
        # 1 - We decode the chromosome
        t0 = x[0]
        T = x[1]
        m_f = x[2]
        # We extract the number of segments for forward and backward
        # propagation
        n_seg = self.__n_seg
        fwd_seg = self.__fwd_seg
        bwd_seg = self.__bwd_seg
        # We extract information on the spacecraft
        m_i = self.__sc.mass
        max_thrust = self.__sc.thrust
        isp = self.__sc.isp
        veff = isp * pk.G0
        # And on the leg
        throttles = [x[3 + 3 * i: 6 + 3 * i] for i in range(n_seg)]
        # Return lists
        n_points_fwd = fwd_seg + 1
        n_points_bwd = bwd_seg + 1
        rfwd = [[0.0] * 3] * (n_points_fwd)
        vfwd = [[0.0] * 3] * (n_points_fwd)
        mfwd = [0.0] * (n_points_fwd)
        ufwd = [[0.0] * 3] * (fwd_seg)
        dfwd = [[0.0] * 3] * (fwd_seg)  # distances E/S
        rbwd = [[0.0] * 3] * (n_points_bwd)
        vbwd = [[0.0] * 3] * (n_points_bwd)
        mbwd = [0.0] * (n_points_bwd)
        ubwd = [[0.0] * 3] * (bwd_seg)
        dbwd = [[0.0] * 3] * (bwd_seg)

        # 2 - We compute the initial and final epochs and ephemerides
        t_i = pk.epoch(t0)
        r_i, v_i = self.__earth.eph(t_i)
        if self.__start == 'l1':
            r_i = [r * 0.99 for r in r_i]
            v_i = [v * 0.99 for v in v_i]
        elif self.__start == 'l2':
            r_i = [r * 1.01 for r in r_i]
            v_i = [v * 1.01 for v in v_i]
        t_f = pk.epoch(t0 + T)
        r_f, v_f = self.target.eph(t_f)

        # 3 - Forward propagation
        fwd_grid = t0 + T * self.__fwd_grid  # days
        fwd_dt = T * self.__fwd_dt  # seconds
        # Initial conditions
        rfwd[0] = r_i
        vfwd[0] = v_i
        mfwd[0] = m_i
        # Propagate
        for i, t in enumerate(throttles[:fwd_seg]):
            if self.__sep:
                r = math.sqrt(rfwd[i][0]**2 + rfwd[i][1]
                              ** 2 + rfwd[i][2]**2) / pk.AU
                max_thrust, isp = self._sep_model(r)
                veff = isp * pk.G0
            ufwd[i] = [max_thrust * thr for thr in t]
            if self.__earth_gravity:
                r_E, v_E = self.__earth.eph(pk.epoch(fwd_grid[i]))
                dfwd[i] = [a - b for a, b in zip(r_E, rfwd[i])]
                r3 = sum([r**2 for r in dfwd[i]])**(3 / 2)
                disturbance = [mfwd[i] * pk.MU_EARTH /
                               r3 * ri for ri in dfwd[i]]
                rfwd[i + 1], vfwd[i + 1], mfwd[i + 1] = pk.propagate_taylor_disturbance(
                    rfwd[i], vfwd[i], mfwd[i], ufwd[i], disturbance, fwd_dt[i], pk.MU_SUN, veff, -10, -10)
            else:
                rfwd[i + 1], vfwd[i + 1], mfwd[i + 1] = pk.propagate_taylor(
                    rfwd[i], vfwd[i], mfwd[i], ufwd[i], fwd_dt[i], pk.MU_SUN, veff, -10, -10)

        # 4 - Backward propagation
        bwd_grid = t0 + T * self.__bwd_grid  # days
        bwd_dt = T * self.__bwd_dt  # seconds
        # Final conditions
        rbwd[-1] = r_f
        vbwd[-1] = v_f
        mbwd[-1] = m_f
        # Propagate
        for i, t in enumerate(throttles[-1:-bwd_seg - 1:-1]):
            if self.__sep:
                r = math.sqrt(rbwd[-i - 1][0]**2 + rbwd[-i - 1]
                              [1]**2 + rbwd[-i - 1][2]**2) / pk.AU
                max_thrust, isp = self._sep_model(r)
                veff = isp * pk.G0
            ubwd[-i - 1] = [max_thrust * thr for thr in t]
            if self.__earth_gravity:
                r_E, v_E = self.__earth.eph(pk.epoch(bwd_grid[-i - 1]))
                dbwd[-i - 1] = [a - b for a, b in zip(r_E, rbwd[-i - 1])]
                r3 = sum([r**2 for r in dbwd[-i - 1]])**(3 / 2)
                disturbance = [mfwd[i] * pk.MU_EARTH /
                               r3 * ri for ri in dbwd[-i - 1]]
                rbwd[-i - 2], vbwd[-i - 2], mbwd[-i - 2] = pk.propagate_taylor_disturbance(
                    rbwd[-i - 1], vbwd[-i - 1], mbwd[-i - 1], ubwd[-i - 1], disturbance, -bwd_dt[-i - 1], pk.MU_SUN, veff, -10, -10)
            else:
                rbwd[-i - 2], vbwd[-i - 2], mbwd[-i - 2] = pk.propagate_taylor(
                    rbwd[-i - 1], vbwd[-i - 1], mbwd[-i - 1], ubwd[-i - 1], -bwd_dt[-i - 1], pk.MU_SUN, veff, -10, -10)

        return rfwd, rbwd, vfwd, vbwd, mfwd, mbwd, ufwd, ubwd, fwd_dt, bwd_dt, dfwd, dbwd

    # Visualizes the trajectory
    def plot_traj(self, x, units=pk.AU, plot_segments=True, plot_thrusts=False, axes=None):
        """
        ax = prob.plot_traj(self, x, units=AU, plot_segments=True, plot_thrusts=False, axes=None)

        Args:
            - x (``list``, ``tuple``, ``numpy.ndarray``): Decision chromosome, e.g. (``pygmo.population.champion_x``).
            - units (``float``): the length unit to be used in the plot
            - plot_segments (``bool``): when True plots also the segments boundaries
            - plot_thrusts (``bool``): when True plots also the thrust vectors

        Returns:
            matplotlib.axes: axes where to plot

        Visualizes the the trajectory in a 3D plot
        """

        if not len(x) == len(self.get_bounds()[0]):
            raise ValueError("Invalid length of the decision vector x")

        import matplotlib as mpl
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        # Creating the axes if necessary
        if axes is None:
            mpl.rcParams['legend.fontsize'] = 10
            fig = plt.figure()
            axes = fig.gca(projection='3d')

        n_seg = self.__n_seg
        fwd_seg = self.__fwd_seg
        bwd_seg = self.__bwd_seg
        t0 = x[0]
        T = x[1]
        isp = self.__sc.isp
        veff = isp * pk.G0
        fwd_grid = t0 + T * self.__fwd_grid  # days
        bwd_grid = t0 + T * self.__bwd_grid  # days

        throttles = [x[3 + 3 * i: 6 + 3 * i] for i in range(n_seg)]
        alphas = [min(1., np.linalg.norm(t)) for t in throttles]

        times = np.concatenate((fwd_grid, bwd_grid))

        rfwd, rbwd, vfwd, vbwd, mfwd, mbwd, ufwd, ubwd, fwd_dt, bwd_dt, dfwd, dbwd = self._propagate(
            x)

        # Plotting the Sun, the Earth and the target
        axes.scatter([0], [0], [0], color='y')
        pk.orbit_plots.plot_planet(self.__earth, pk.epoch(
            t0), units=units, legend=True, color=(0.7, 0.7, 1), axes=axes)
        pk.orbit_plots.plot_planet(self.target, pk.epoch(
            t0 + T), units=units, legend=True, color=(0.7, 0.7, 1), axes=axes)

        # Forward propagation
        xfwd = [0.0] * (fwd_seg + 1)
        yfwd = [0.0] * (fwd_seg + 1)
        zfwd = [0.0] * (fwd_seg + 1)
        xfwd[0] = rfwd[0][0] / units
        yfwd[0] = rfwd[0][1] / units
        zfwd[0] = rfwd[0][2] / units

        for i in range(fwd_seg):
            if self.__sep:
                r = math.sqrt(rfwd[i][0]**2 + rfwd[i][1]
                              ** 2 + rfwd[i][2]**2) / pk.AU
                _, isp = self._sep_model(r)
                veff = isp * pk.G0
            if self.__earth_gravity:
                r3 = sum([r**2 for r in dfwd[i]])**(3 / 2)
                disturbance = [mfwd[i] * pk.MU_EARTH /
                               r3 * ri for ri in dfwd[i]]
                pk.orbit_plots.plot_taylor_disturbance(rfwd[i], vfwd[i], mfwd[i], ufwd[i], disturbance, fwd_dt[
                                                       i], pk.MU_SUN, veff, N=10, units=units, color=(alphas[i], 0, 1 - alphas[i]), axes=axes)
            else:
                pk.orbit_plots.plot_taylor(rfwd[i], vfwd[i], mfwd[i], ufwd[i], fwd_dt[
                                           i], pk.MU_SUN, veff, N=10, units=units, color=(alphas[i], 0, 1 - alphas[i]), axes=axes)
            xfwd[i + 1] = rfwd[i + 1][0] / units
            yfwd[i + 1] = rfwd[i + 1][1] / units
            zfwd[i + 1] = rfwd[i + 1][2] / units
        if plot_segments:
            axes.scatter(xfwd[:-1], yfwd[:-1], zfwd[:-1],
                         label='nodes', marker='o', s=5, c='k')

        # Backward propagation
        xbwd = [0.0] * (bwd_seg + 1)
        ybwd = [0.0] * (bwd_seg + 1)
        zbwd = [0.0] * (bwd_seg + 1)
        xbwd[-1] = rbwd[-1][0] / units
        ybwd[-1] = rbwd[-1][1] / units
        zbwd[-1] = rbwd[-1][2] / units

        for i in range(bwd_seg):
            if self.__sep:
                r = math.sqrt(rbwd[-i - 1][0]**2 + rbwd[-i - 1]
                              [1]**2 + rbwd[-i - 1][2]**2) / pk.AU
                _, isp = self._sep_model(r)
                veff = isp * pk.G0
            if self.__earth_gravity:
                r3 = sum([r**2 for r in dbwd[-i - 1]])**(3 / 2)
                disturbance = [mfwd[i] * pk.MU_EARTH /
                               r3 * ri for ri in dbwd[-i - 1]]
                pk.orbit_plots.plot_taylor_disturbance(rbwd[-i - 1], vbwd[-i - 1], mbwd[-i - 1], ubwd[-i - 1], disturbance, -bwd_dt[
                                                       -i - 1], pk.MU_SUN, veff, N=10, units=units, color=(alphas[-i - 1], 0, 1 - alphas[-i - 1]), axes=axes)
            else:
                pk.orbit_plots.plot_taylor(rbwd[-i - 1], vbwd[-i - 1], mbwd[-i - 1], ubwd[-i - 1], -bwd_dt[-i - 1],
                                           pk.MU_SUN, veff, N=10, units=units, color=(alphas[-i - 1], 0, 1 - alphas[-i - 1]), axes=axes)
            xbwd[-i - 2] = rbwd[-i - 2][0] / units
            ybwd[-i - 2] = rbwd[-i - 2][1] / units
            zbwd[-i - 2] = rbwd[-i - 2][2] / units
        if plot_segments:
            axes.scatter(xbwd[1:], ybwd[1:], zbwd[1:], marker='o', s=5, c='k')

        # Plotting the thrust vectors
        if plot_thrusts:
            throttles = np.array(throttles)
            xlim = axes.get_xlim()
            xrange = xlim[1] - xlim[0]
            ylim = axes.get_ylim()
            yrange = ylim[1] - ylim[0]
            zlim = axes.get_zlim()
            zrange = zlim[1] - zlim[0]

            scale = 0.1

            throttles[:, 0] *= xrange
            throttles[:, 1] *= yrange
            throttles[:, 2] *= zrange

            throttles *= scale

            for (x, y, z, t) in zip(xfwd[:-1] + xbwd[:-1], yfwd[:-1] + ybwd[:-1], zfwd[:-1] + zbwd[:-1], throttles):
                axes.plot([x, x + t[0]], [y, y + t[1]], [z, z + t[2]], c='g')

        return axes

    def plot_dists_thrust(self, x, axes=None):
        """
        axes = prob.plot_dists_thrust(self, x, axes=None)

        Args:
            - x (``list``, ``tuple``, ``numpy.ndarray``): Decision chromosome, e.g. (``pygmo.population.champion_x``).

        Returns:
            matplotlib.axes: axes where to plot

        Plots the distance of the spacecraft from the Earth/Sun and the thrust profile
        """

        if not len(x) == len(self.get_bounds()[0]):
            raise ValueError("Invalid length of the decision vector x")

        import matplotlib as mpl
        import matplotlib.pyplot as plt

        # Creating the axes if necessary
        if axes is None:
            mpl.rcParams['legend.fontsize'] = 10
            fig = plt.figure()
            axes = fig.add_subplot(111)

        n_seg = self.__n_seg
        fwd_seg = self.__fwd_seg
        bwd_seg = self.__bwd_seg
        t0 = x[0]
        T = x[1]
        fwd_grid = t0 + T * self.__fwd_grid  # days
        bwd_grid = t0 + T * self.__bwd_grid  # days

        throttles = [np.linalg.norm(x[3 + 3 * i: 6 + 3 * i])
                     for i in range(n_seg)]

        dist_earth = [0.0] * (n_seg + 2)  # distances spacecraft - Earth
        dist_sun = [0.0] * (n_seg + 2)  # distances spacecraft - Sun
        times = np.concatenate((fwd_grid, bwd_grid))

        rfwd, rbwd, vfwd, vbwd, mfwd, mbwd, ufwd, ubwd, fwd_dt, bwd_dt, _, _ = self._propagate(
            x)

        # Forward propagation
        xfwd = [0.0] * (fwd_seg + 1)
        yfwd = [0.0] * (fwd_seg + 1)
        zfwd = [0.0] * (fwd_seg + 1)
        xfwd[0] = rfwd[0][0] / pk.AU
        yfwd[0] = rfwd[0][1] / pk.AU
        zfwd[0] = rfwd[0][2] / pk.AU
        r_E = [ri / pk.AU for ri in self.__earth.eph(pk.epoch(fwd_grid[0]))[0]]
        dist_earth[0] = np.linalg.norm(
            [r_E[0] - xfwd[0], r_E[1] - yfwd[0], r_E[2] - zfwd[0]])
        dist_sun[0] = np.linalg.norm([xfwd[0], yfwd[0], zfwd[0]])

        for i in range(fwd_seg):
            xfwd[i + 1] = rfwd[i + 1][0] / pk.AU
            yfwd[i + 1] = rfwd[i + 1][1] / pk.AU
            zfwd[i + 1] = rfwd[i + 1][2] / pk.AU
            r_E = [
                ri / pk.AU for ri in self.__earth.eph(pk.epoch(fwd_grid[i + 1]))[0]]
            dist_earth[
                i + 1] = np.linalg.norm([r_E[0] - xfwd[i + 1], r_E[1] - yfwd[i + 1], r_E[2] - zfwd[i + 1]])
            dist_sun[
                i + 1] = np.linalg.norm([xfwd[i + 1], yfwd[i + 1], zfwd[i + 1]])

        # Backward propagation
        xbwd = [0.0] * (bwd_seg + 1)
        ybwd = [0.0] * (bwd_seg + 1)
        zbwd = [0.0] * (bwd_seg + 1)
        xbwd[-1] = rbwd[-1][0] / pk.AU
        ybwd[-1] = rbwd[-1][1] / pk.AU
        zbwd[-1] = rbwd[-1][2] / pk.AU
        r_E = [
            ri / pk.AU for ri in self.__earth.eph(pk.epoch(bwd_grid[-1]))[0]]
        dist_earth[-1] = np.linalg.norm([r_E[0] - xbwd[-1],
                                         r_E[1] - ybwd[-1], r_E[2] - zbwd[-1]])
        dist_sun[-1] = np.linalg.norm([xbwd[-1], ybwd[-1], zbwd[-1]])

        for i in range(bwd_seg):
            xbwd[-i - 2] = rbwd[-i - 2][0] / pk.AU
            ybwd[-i - 2] = rbwd[-i - 2][1] / pk.AU
            zbwd[-i - 2] = rbwd[-i - 2][2] / pk.AU
            r_E = [
                ri / pk.AU for ri in self.__earth.eph(pk.epoch(bwd_grid[-i - 2]))[0]]
            dist_earth[-i - 2] = np.linalg.norm(
                [r_E[0] - xbwd[-i - 2], r_E[1] - ybwd[-i - 2], r_E[2] - zbwd[-i - 2]])
            dist_sun[-i -
                     2] = np.linalg.norm([xbwd[-i - 2], ybwd[-i - 2], zbwd[-i - 2]])

        axes.set_xlabel("t [mjd2000]")
        # Plot Earth distance
        axes.plot(times, dist_earth, c='b', label="sc-Earth")
        # Plot Sun distance
        axes.plot(times, dist_sun, c='y', label="sc-Sun")
        axes.set_ylabel("distance [AU]", color='k')
        axes.set_ylim(bottom=0.)
        axes.tick_params('y', colors='k')
        axes.legend(loc=2)
        # draw threshold where Earth gravity equals 0.1*Tmax
        if self.__earth_gravity:
            axes.axhline(y=np.sqrt(pk.MU_EARTH * self.__sc.mass / (self.__sc.thrust * 0.1)
                                   ) / pk.AU, c='g', ls=":", lw=1, label="earth_g = 0.1*Tmax")
        # Plot thrust profile
        axes = axes.twinx()
        if self.__sep:
            max_thrust = self.__sc.thrust
            thrusts = np.linalg.norm(
                np.array(ufwd + ubwd), axis=1) / max_thrust
            # plot maximum thrust achievable at that distance from the Sun
            distsSun = dist_sun[:fwd_seg] + \
                dist_sun[-bwd_seg:] + [dist_sun[-1]]
            Tmaxs = [self._sep_model(d)[0] / max_thrust for d in distsSun]
            axes.step(np.concatenate(
                (fwd_grid, bwd_grid[1:])), Tmaxs, where="post", c='lightgray', linestyle=':')
        else:
            thrusts = throttles.copy()
        # duplicate the last for plotting
        thrusts = np.append(thrusts, thrusts[-1])
        axes.step(np.concatenate(
            (fwd_grid, bwd_grid[1:])), thrusts, where="post", c='r', linestyle='--')
        axes.set_ylabel("T/Tmax$_{1AU}$", color='r')
        axes.tick_params('y', colors='r')
        axes.set_xlim([times[0], times[-1]])
        axes.set_ylim([0, max(thrusts) + 0.2])

        return axes

    def double_segments(self, x):
        """
        new_prob, new_x = prob.double_segments(self,x)

        Args:
            - x (``list``, ``tuple``, ``numpy.ndarray``): Decision chromosome, e.g. (``pygmo.population.champion_x``).

        Returns:
            - the new udp having twice the segments
            - list: the new chromosome to be used as initial guess

        Returns the decision vector encoding a low trust trajectory having double the number of segments with respect to x
        and a 'similar' throttle history. In case high fidelity is True, and x is a feasible trajectory, the returned decision vector
        also encodes a feasible trajectory that can be further optimized
        Returns the problem and the decision vector encoding a low-thrust trajectory having double the number of
        segments with respect to x and the same throttle history. If x is a feasible trajectory, the new chromosome is "slightly
        unfeasible", due to the new refined Earth's gravity and/or SEP thrust that are now computed in the 2 halves of each segment.
        """

        if not len(x) == len(self.get_bounds()[0]):
            raise ValueError("Invalid length of the decision vector x")

        new_x = np.append(x[:3], np.repeat(x[3:].reshape((-1, 3)), 2, axis=0))

        new_prob = lt_margo(
            target=self.target,
            n_seg=2 * self.__n_seg,
            grid_type=self.__grid_type,
            t0=[pk.epoch(self.__lb[0]), pk.epoch(self.__ub[0])],
            tof=[self.__lb[1], self.__ub[1]],
            m0=self.__sc.mass,
            Tmax=self.__sc.thrust,
            Isp=self.__sc.isp,
            earth_gravity=self.__earth_gravity,
            sep=self.__sep,
            start=self.__start
        )

        return new_prob, new_x

    def pretty(self, x):
        """
        prob.pretty(x)

        Args:
            - x (``list``, ``tuple``, ``numpy.ndarray``): Decision chromosome, e.g. (``pygmo.population.champion_x``).

        Prints human readable information on the trajectory represented by the decision vector x
        """

        if not len(x) == len(self.get_bounds()[0]):
            raise ValueError("Invalid length of the decision vector x")

        n_seg = self.__n_seg
        m_i = self.__sc.mass
        t0 = x[0]
        T = x[1]
        m_f = x[2]
        thrusts = [np.linalg.norm(x[3 + 3 * i: 6 + 3 * i])
                   for i in range(n_seg)]

        tf = t0 + T
        mP = m_i - m_f
        deltaV = self.__sc.isp * pk.G0 * np.log(m_i / m_f)

        dt = np.append(self.__fwd_dt, self.__bwd_dt) * T / pk.DAY2SEC
        time_thrusts_on = sum(dt[i] for i in range(
            len(thrusts)) if thrusts[i] > 0.1)

        print("Departure:", pk.epoch(t0), "(", t0, "mjd2000 )")
        print("Time of flight:", T, "days")
        print("Arrival:", pk.epoch(tf), "(", tf, "mjd2000 )")
        print("Delta-v:", deltaV, "m/s")
        print("Propellant consumption:", mP, "kg")
        print("Thrust-on time:", time_thrusts_on, "days")
